from SBI.beans.IndexedNum import IndexedNum
import time
import re


def process_HEADER_line(line):
    '''
    COLUMNS  DATA TYPE    FIELD           DEFINITION
    ---------------------------------------------------------------------------
     1 -  6  Record name  "HEADER"
    11 - 50  String(40)   classification  Classifies the molecule(s)
    51 - 59  Date         depDate         Deposition date.
                                          his is the date the coordinates were
                                          received by the PDB
    63 - 66  IDcode       idCode          This identifier is unique in the PDB
    '''
    data = []
    data.append(line[10:50].strip())
    inf, outf = "%d-%b-%y", "%Y-%m-%d"
    if line[50:60].strip() != '':
        tme = time.strftime(outf, time.strptime(line[50:60].strip(), inf))
    else:
        tme = ''
    data.append(tme)
    data.append(line[62:67].strip())
    return data


def process_TITLE_line(line):
    '''
    COLUMNS  DATA TYPE     FIELD         DEFINITION
    --------------------------------------------------------------------------
     1 -  6  Record name   "TITLE "
     9 - 10  Continuation  continuation  Concatenation of multiple records.
    11 - 70  String        title         Title of the experiment.
    '''
    return line[10:].strip()


def process_EXPERIMENT_line(line):
    '''
    COLUMNS       DATA TYPE      FIELD         DEFINITION
    -------------------------------------------------------
     1 -  6       Record name    "EXPDTA"
     9 - 10       Continuation   continuation  Allows concatenation
                                               of multiple records
    11 - 70       SList          technique     The experimental technique(s)
                                               with optional comment describing
                                               the sample or experiment.
    '''
    return line[10:].split(';')[0].strip()


def process_RESOLUTION_line(line):
    '''
    COLUMNS    DATA TYPE      FIELD              DEFINITION
    -------------------------------------------------------
     1 - 6     Record name    "REMARK"
    10         LString(1)     "2"
    12 - 22    LString(11)    "RESOLUTION."
    23 - 27    Real(5.2)      resolution         Resolution.
    29 - 38    LString(10)    "ANGSTROMS."
    '''
    data = line[23:30].strip().split()[0]
    return None if data == 'NULL' or data == 'NOT' else data


def process_RFACTOR_line(line):
    if line.strip().endswith('NULL'):
        return None
    return line.split(':')[-1].strip()


def process_FREER_line(line):
    if line.strip().endswith('NULL'):
        return None
    return line.split(':')[-1].strip()


def process_SUPERSEEDED_line(line):
    '''
    COLUMNS        DATA TYPE     FIELD         DEFINITION
    -------------------------------------------------------------------------
     1 -  6        Record name   "SPRSDE"
     9 - 10        Continuation  continuation  Allows for multiple ID codes.
    12 - 20        Date          sprsdeDate    Date it superseded others
    22 - 25        IDcode        idCode        ID code of this entry.
    32 - 35        IDcode        sIdCode       ID code of a superseded entry.
    37 - 40        IDcode        sIdCode       ID code of a superseded entry.
    42 - 45        IDcode        sIdCode       ID code of a superseded entry.
    47 - 50        IDcode        sIdCode       ID code of a superseded entry.
    52 - 55        IDcode        sIdCode       ID code of a superseded entry.
    57 - 60        IDcode        sIdCode       ID code of a superseded entry.
    62 - 65        IDcode        sIdCode       ID code of a superseded entry.
    67 - 70        IDcode        sIdCode       ID code of a superseded entry.
    72 - 75        IDcode        sIdCode       ID code of a superseded entry.
    '''
    return line[10:].strip().split()[2:]


def process_MOLKEY(line):
    key = line.split(':')[-1].strip()
    return re.sub(';', '', key)


def process_COMPND_line(line):
    '''
    COLUMNS  DATA TYPE       FIELD         DEFINITION
    --------------------------------------------------------------------------
     1 -  6  Record name     "COMPND"
     8 - 10  Continuation    continuation  Concatenation of multiple records.
    11 - 80  Specification   compound      Description of molecular components.
             list
    '''
    return line[10:].strip()


def process_SOURCE_line(line):
    '''
    COLUMNS DATA  TYPE     FIELD          DEFINITION
    -------------------------------------------------------------------------
     1 -  6 Record name    "SOURCE"
     8 - 10 Continuation   continuation   Concatenation of multiple records.
    11 - 79 Specification  srcName        Identifies the source of the
            List                          macromolecule in a token.
    '''
    return line[10:].strip()


def process_KEYWRD_line(line):
    '''
        COLUMNS       DATA  TYPE     FIELD         DEFINITION
    -------------------------------------------------------------------
     1 -  6       Record name    "KEYWDS"
     9 - 10       Continuation   continuation  Concatenation of records.
    11 - 79       List           keywds        Comma-separated list
    '''
    return line[10:].strip()


def process_SITE_IDENTIFIER_line(line):
    '''
             1         2         3         4         5         6
    123456789012345678901234567890123456789012345678901234567890
    REMARK 800
    REMARK 800 SITE
    REMARK 800 SITE_IDENTIFIER: FREE TEXT GOES HERE.
    REMARK 800 EVIDENCE_CODE: (AUTHOR or SOFTWARE or UNKNOWN)
    REMARK 800 SITE_DESCRIPTION: FREE TEXT GOES HERE.
    '''
    return line[10:].split(':')[1].strip()


def process_REMARK800_line(line):
    return line[10:].strip()


def process_SITE_lines(line):
    '''
             1         2         3         4         5         6
    123456789012345678901234567890123456789012345678901234567890
    SITE     1 AC1  3 HIS A  94  HIS A  96  HIS A 119
    SITE     1 AC2  5 ASN A  62  GLY A  63  HIS A  64  HOH A 328
    SITE     2 AC2  5 HOH A 634
    SITE     1 AC3  5 GLN A 136  GLN A 137  PRO A 138  GLU A 205
    SITE     2 AC3  5 CYS A 206
    SITE     1 AC4 11 HIS A  64  HIS A  94  HIS A  96  HIS A 119
    SITE     2 AC4 11 LEU A 198  THR A 199  THR A 200  TRP A 209
    SITE     3 AC4 11 HOH A 572  HOH A 582  HOH A 635
    '''
    data   = [line[10:14].strip()]
    line   = line[18:].rstrip()
    i = 0
    while i + 9 <= len(line):
        resi = line[i:i + 11]
        rest = resi[:3].strip()
        resp = resi[3:].strip()
        data.append(tuple([rest, resp]))
        i += 11
    return data


def process_HETEROATOM_lines(line):
    '''
    There are 3 lines that define an heteroatom:

    COLUMNS     DATA TYPE     FIELD         DEFINITION
    ------------------------------------------------------
     1 -  6     Record name   "HET      "
     8 - 10     LString(3)    hetID         Het identifier, right-justified.
    13          Character     ChainID       Chain identifier.
    14 - 17     Integer       seqNum        Sequence number.
    18          AChar         iCode         Insertion code.
    21 - 25     Integer       numHetAtoms   Number of HETATM records for the
                                            group present in the entry.
    31 - 70     String        text          Text describing Het group.

             1         2         3         4         5         6         7
    1234567890123456789012345678901234567890123456789012345678901234567890
    HET    TRS    975       8
    HET    STA  I   4      25     PART_OF: HIV INHIBITOR;

    HET    FUC  Y   1      10     PART_OF: NONOATE COMPLEX; L-FUCOSE
    HET    GAL  Y   2      11     PART_OF: NONOATE COMPLEX
    HET    NAG  Y   3      15
    HET    FUC  Y   4      10
    HET    NON  Y   5      12

    HET    UNX  A 161       1     PSEUDO CARBON ATOM OF UNKNOWN LIGAND
    HET    UNX  A 162       1     PSEUDO CARBON ATOM OF UNKNOWN LIGAND
    HET    UNX  A 163       1     PSEUDO CARBON ATOM OF UNKNOWN LIGAND

    COLUMNS     DATA TYPE       FIELD          DEFINITION
    -------------------------------------------------------
     1 -  6     Record name     "HETNAM"
     9 - 10     Continuation    continuation   Allows concatenation of
                                               multiple records.
    12 - 14     LString(3)      hetID          Het identifier,
                                               right-justified.
    16 - 70     String          text           Chemical name.

             1         2         3         4         5         6         7
    1234567890123456789012345678901234567890123456789012345678901234567890
    HETNAM     GLC GLUCOSE
    HETNAM     SAD BETA-METHYLENE SELENAZOLE-4-CARBOXAMIDE ADENINE
    HETNAM  2  SAD DINUCLEOTIDE

    HETNAM     UNX UNKNOWN ATOM OR ION
    HETNAM     UNL UNKNOWN LIGAND

    HETNAM     CYE 45-(3-AMINOPROPYL)-5,11,22,28,34-PENTAMETHYL-3,9,15,
    HETNAM   2 CYE  20,26,32,38,43-OCTAOXO-2,5,8,14,19,22,25,28,31,34,37,
    HETNAM   3 CYE  42,45,48-TETRADECAAZA-11-AZONIAHEPTACYCLO[42.2.1.1~4,
    HETNAM   4 CYE  7~.1~10,13~.1~21,24~.1~27,30~.1~33,36~]DOPENTACONTA-
    HETNAM   5 CYE  1(46),4(52),6,10(51),12,21(50),23,27(49),29,33(48),35,
    HETNAM   6 CYE  44(47)-DODECAENE

    COLUMNS     DATA TYPE       FIELD          DEFINITION
    -------------------------------------------------------
     1 -  6     Record name     "FORMUL"
     9 - 10     Integer         compNum        Component number.
    13 - 15     LString(3)      hetID          Het identifier.
    17 - 18     Integer         continuation   Continuation number.
    19          Character       asterisk       "*" for water.
    20 - 70     String          text           Chemical formula.

             1         2         3         4         5         6         7
    1234567890123456789012345678901234567890123456789012345678901234567890
    FORMUL   2  SO4    2(O4 S1 2-)
    FORMUL   3  GLC    C6 H12 O6
    FORMUL   3  FOL    2(C19 H17 N7 O6 2-)
    FORMUL   4   CL    2(CL1 1-)
    FORMUL   5   CA    CA1 2+

    FORMUL   1  ACE    C2 H3 O1
    FORMUL   2  ACE    C2 H3 O1

    FORMUL   8  HOH   *463(H2 O1)
    '''
    data = []
    if line.startswith("HET   "):
        data.append(line[5:].strip().split()[0])
        data.append(line[11:17].strip())
    elif line.startswith("HETNAM"):
        data.append(line[10:14].strip())
        data.append(line[15:].strip())
    elif line.startswith("FORMUL"):
        data.append(line[11:15].strip())
        formula = re.sub(r'^\d+\(', '', line[16:].strip())
        formula = re.sub(r'\)\Z', '', formula)
        data.append(formula)
    return data


def process_DBREF_line(line):
    '''
    The line format is such as:

    COLUMNS    DATA TYPE    FIELD          DEFINITION
    ----------------------------------------------------------------
     1 - 6     Record name  "DBREF "
     8 - 11    IDcode       idCode         ID code of this entry.
    13         Character    chainID        Chain identifier.
    15 - 18    Integer      seqBegin       Initial sequence number
                                           of the PDB sequence segment.
    19         AChar        insertBegin    Initial insertion code
                                           of the PDB sequence segment.
    21 - 24   Integer       seqEnd         Ending sequence number
                                           of the PDB sequence segment.
    25         AChar        insertEnd      Ending insertion code
                                           of the PDB sequence segment.
    27 - 32    LString      database       Sequence database name.
    34 - 41    LString      dbAccession    Sequence database accession code.
    43 - 54    LString      dbIdCode       Sequence database
                                           identification code.
    56 - 60    Integer      dbseqBegin     Initial sequence number of the
                                           database seqment.
    61         AChar        idbnsBeg       Insertion code of initial residue
                                           of the segment, if PDB is the
                                           reference.
    63 - 67    Integer      dbseqEnd       Ending sequence number of the
                                           database segment.
    68         AChar        dbinsEnd       Insertion code of the ending
                                           residue of the segment, if PDB is
                                           the reference.

    For example:

             1         2         3         4         5         6         7
    1234567890123456789012345678901234567890123456789012345678901234567890
    DBREF  1ABC B    1B   36  PDB    1ABC     1ABC             1B    36

    DBREF  3AKY      3   220  SWS    P07170   KAD1_YEAST       5    222

    DBREF  1HAN      2   288  GB     397884   X66122           1    287

    DBREF  3HSV A    1    92  SWS    P22121   HSF_KLULA      193    284
    DBREF  3HSV B    1    92  SWS    P22121   HSF_KLULA      193    284

    DBREF  1ARL      1   307  SWS    P00730   CBPA_BOVIN     111    417

    DBREF  249D A    1    12  NDB    BDL070   BDL070           1     12
    DBREF  249D B   13    24  NDB    BDL070   BDL070          13     24
    DBREF  249D C   26    36  NDB    BDL070   BDL070          26     36
    DBREF  249D D   37    48  NDB    BDL070   BDL070          37     48

    Detailed info @ http://www.wwpdb.org/documentation/format23/sect3.html
    '''
    data = []
    data.append(line[7:11].strip())
    data.append(fix_chain(line[12:13].strip()))
    data.append(IndexedNum(line[14:19].strip()))
    data.append(IndexedNum(line[20:25].strip()))
    data.append(line[25:29].strip())
    data.append(line[30:41].strip())
    return data


def process_SECONDARYSTRUCTURE_lines(line):
    '''
    COLUMNS  DATA  TYPE     FIELD         DEFINITION
    ---------------------------------------------------------------------------
     1 -  6  Record name    "HELIX "
     8 - 10  Integer        serNum        Serial number of the helix. Starts
                                          at 1  and increases incrementally.
    12 - 14  LString(3)     helixID       Helix  identifier.
    16 - 18  Residue name   initResName   Name of the initial residue.
    20       Character      initChainID   Chain identifier
    22 - 25  Integer        initSeqNum    Number of the initial residue.
    26       AChar          initICode     Code of the initial residue.
    28 - 30  Residue  name  endResName    Name of the terminal residue
    32       Character      endChainID    Chain identifier
    34 - 37  Integer        endSeqNum     Number of the terminal residue.
    38       AChar          endICode      Code of the terminal residue.
    39 - 40  Integer        helixClass    Helix class (see below).
    41 - 70  String         comment       Comment about this helix.
    72 - 76  Integer        length        Length of this helix.

    COLUMNS  DATA  TYPE     FIELD          DEFINITION
    --------------------------------------------------------------------------
     1 -  6  Record name   "SHEET "
     8 - 10  Integer       strand         Strand number
    12 - 14  LString(3)    sheetID        Sheet  identifier.
    15 - 16  Integer       numStrands     Number  of strands in sheet.
    18 - 20  Residue name  initResName    Residue  name of initial residue.
    22       Character     initChainID    Chain identifier of initial residue
                                          in strand.
    23 - 26  Integer       initSeqNum     Sequence number of initial residue
                                          in strand.
    27       AChar         initICode      Insertion code of initial residue
                                          in  strand.
    29 - 31  Residue name  endResName     Residue name of terminal residue.
    33       Character     endChainID     Chain identifier of terminal residue.
    34 - 37  Integer       endSeqNum      Sequence number of terminal residue.
    38       AChar         endICode       Insertion code of terminal residue.
    39 - 40  Integer       sense          Sense of strand.
                                          0 if first strand, 1 if  parallel
                                          and -1 if anti-parallel.
    42 - 45  Atom          curAtom        Atom name in current strand.
    46 - 48  Residue name  curResName     Residue name in current strand
    50       Character     curChainId     Chain identifier in
                                          current strand.
    51 - 54  Integer       curResSeq      Residue sequence number
                                          in current strand.
    55       AChar         curICode       Insertion code in
                                          current strand.
    57 - 60  Atom          prevAtom       Atom name in previous strand.
    61 - 63  Residue name  prevResName    Registration.  Residue name in
                                          previous strand.
    65       Character     prevChainId    Registration.  Chain identifier in
                                          previous  strand.
    66 - 69  Integer       prevResSeq     Registration. Residue sequence number
                                          in previous strand.
    70       AChar         prevICode      Registration.  Insertion code in
                                          previous strand.
    '''
    return line.strip()


def process_MATRIX_lines(line):
    '''
    COLUMNS        DATA  TYPE    FIELD         DEFINITION
    --------------------------------------------------------------------------
     1 -  6        Record name   "MTRIXn"      n=1, 2, or 3
     8 - 10        Integer       serial        Serial number.
    11 - 20        Real(10.6)    m[n][1]       Mn1
    21 - 30        Real(10.6)    m[n][2]       Mn2
    31 - 40        Real(10.6)    m[n][3]       Mn3
    46 - 55        Real(10.5)    v[n]          Vn
    60             Integer       iGiven
    '''
    # Matrix data can and MUST be gathered not by position
    # but by spliting the line
    # Why? 1M4X has more than 1000 matrices and shifts the position
    #      1QKP is wrongly formated and shifts the position
    matrix_value    = re.split('\s+', line.strip())
    matrix_value    = matrix_value[2:]
    matrix_value[0] = matrix_value[0][-1]
    matrix_value.pop(1)
    return matrix_value

def process_REMARK350_lines(line):
    return  line.strip().split(':')[-1]

def fix_chain(chain):
    '''
    Sometimes old PDBs have no chain, transform to chain A
    '''
    return 'A' if chain.strip() == '' else chain.strip()



"""
BlastHit

The BlastHit object stores all the data related to a blast XML output hit.

That means the data inside each <Hit> tag of the blast XML file which corresponds to a single blast match to the query.

WARNING!: It has to be a blastpgp from a single sequence; multi-fasta blasts are not compatible
          with xml.dom.minidom parsing capabilities.

                                   ############################################
                                        jbonet @ boliva's lab        2011
"""
import re
from collections import Counter

from SBI.sequence.alignment import SeqAli


class BlastHit(SeqAli):
    '''
    A sequence alignment that stores the data related to a single query-hit
    relation from a Blast output.
    All the query-hit ({BlastHit}) pairs are grouped into a {BlastResult}.

    Normally, the object will be created directly through the {BlastParser}

    '''

    _VAST = {'query': 0, 0: 0, 'hit': 1, 1: 1}

    def __init__(self, hit, sequences, sequence_inits, iteration, stats):
        '''

        @param:    sequences
        @pdef:     list with the aligned query and hit sequences (in this order)
                   plus the alignment pattern representation derived from blast
                   (in the last position of the list).
        @ptype:    {List} of {String}

        @param:    sequence_inits
        @pdef:     list of the first position number for the query and hit
                   sequences, in this order.
        @ptype:    {List} of {Integer}

        @param:    iteration
        @pdef:     iteration of the blast output to which the {BlastHit} belongs
        @ptype:    {Integer}

        @param:    stats
        @pdef:     list alignment statistics. Namely: identities, positives, gaps
                   and e-value
        @ptype:    {List} as [{Integer}, {Integer}, {Integer}, {Float}]
        '''
        align_pattern = sequences.pop() if len(sequences) > 2 else None
        super(BlastHit, self).__init__(sequences      = sequences,
                                       sequence_inits = sequence_inits,
                                       identities     = stats[0],
                                       positives      = stats[1],
                                       gaps           = stats[2])

        if align_pattern is not None:
            self.add_alignment_pattern(align_pattern = align_pattern, method = 'blast')

        self._sequenceID   = str(hit[0])        # Hit Name
        self._length       = int(hit[1])        # Hit Length
        self._iteration    = int(iteration)     # Iteration to which the hit belongs
        self._e_value      = stats[3]           # Alignment e-value

        self._used         = False

    ##############
    # ATTRIBUTES #
    ##############
    @property
    def query_seq(self):
        '''
        The aligned query sequence

        @return: {String}
        '''
        return str(self._seq[0].sequence)

    @property
    def ungapped_query_seq(self):
        '''
        The aligned query sequence. without gaps

        @return: {String}
        '''
        return str(re.sub('-', '', self.query_seq))

    @property
    def hit_seq(self):
        '''
        The aligned hit sequence

        @return: {String}
        '''
        return str(self._seq[1].sequence)

    @property
    def ungapped_hit_seq(self):
        '''
        The aligned hit sequence. without gaps

        @return: {String}
        '''
        return str(re.sub('-', '', self.hit_seq))

    @property
    def query_pos(self):
        return self._segment[0]

    @property
    def hit_pos(self):
        return self._segment[1]

    '''ATTRIBUTES'''
    @property
    def sequenceID(self):
        return self._sequenceID

    @property
    def length(self):
        return self._length

    @property
    def iteration(self):
        return self._iteration

    @property
    def evalue(self):
        return float(self._e_value) if self._e_value is not None else None

    @property
    def is_used(self):
        return self._used

    @is_used.setter
    def is_used(self, value):
        self._used = value

    '''MIX ATTRIBUTES'''
    @property
    def query_threshold_coord(self):
        return [self._first_segment_value(0), self._last_segment_value(0)]

    @property
    def query_align_segment_length(self):
        return self._aligned_sequence_length(0)

    @property
    def hit_threshold_coord(self):
        return [self._first_segment_value(1), self._last_segment_value(1)]

    @property
    def hit_align_segment_length(self):
        return self._aligned_sequence_length(1)

    '''OVERWRITTE PARENT GETTERS TO ADAPT TO refseq=["query","hit"]'''

    def get_hit_to_query_coordinate(self, coordinate):
        refseq = self._check_valid_alignment_seq_type(1)
        return super(BlastHit, self).get_respective_coordinate(refseq, int(not refseq), coordinate)

    def get_query_to_hit_coordinate(self, coordinate):
        refseq = self._check_valid_alignment_seq_type(0)
        return super(BlastHit, self).get_respective_coordinate(refseq, int(not refseq), coordinate)

    def get_coverage_of_sequence(self, refseq, fulllength = None, ini = None, end = None):
        """
        METHOD: get_coverage_of_sequence(refseq =["query","hit", fulllength = None, ini = None, end = None))

            __DEVEL_INFO__
            => Overwrittes from parent (SeqAli)

            __EVALUATOR__
            =>Returns coverage ratio (0<->1) of the quey/hit over the full sequence (for query fulllength is required)

            =>If a *range* is given, returns __the sequence coverage for the given range__
             (adjusted to actually aligned regions)

            =>[TIP] To obtain __the sequence coverage for the aligned region over itself__:
                ini = self.query_threshold_coord[0] or self.hit_threshold_coord[0]
                AND
                end = self.query_threshold_coord[1] or self.hit_threshold_coord[1]
        """
        refseq = self._check_valid_alignment_seq_type(refseq)
        if bool(refseq):  # hit
            fulllength = self.length
        else:  # query
            if fulllength is None:
                raise AttributeError('To check the query coverage the query length must be given')

        return super(BlastHit, self).get_coverage_of_sequence(BlastHit._VAST[refseq], fulllength, ini, end)

    def get_min_coverage_of_sequence(self, fulllength):
        return min(self.get_coverage_of_sequence('hit'), self.get_coverage_of_sequence('query', fulllength))

    def get_coverage_of_full_sequence_segment(self, refseq):
        refseq = self._check_valid_alignment_seq_type(refseq)

        if bool(refseq):  # hit
            length = len(self.hit_seq.replace('-', ''))
        else:  # query
            length = len(self.query_seq.replace('-', ''))

        return super(BlastHit, self).get_coverage_of_sequence(BlastHit._VAST[refseq], length)

    def get_min_coverage_of_full_sequence_segment(self):
        return min(self.get_coverage_of_full_sequence_segment('hit'), self.get_coverage_of_full_sequence_segment('query'))

    def get_section_from_sequence_position(self, refseq, ini, end):
        refseq = self._check_valid_alignment_seq_type(refseq)
        return super(BlastHit, self).get_section_from_sequence_position(refseq, ini, end)

    def correct_hit_count(self, new_index):
        if isinstance(self._idx[1], list):
            raise AttributeError('Complex index of hit cannot be modified')

        if not isinstance(new_index, list):
            self.increment_sequence_index(1, new_index) ##
        else:
            self.add_complex_index(1, new_index)

    def correct_query_count(self, new_index):
        if isinstance(self._idx[0], list):
            raise AttributeError('Complex index of querycannot be modified')

        if not isinstance(new_index, list):
            self.increment_sequence_index(0, new_index)
        else:
            self.add_complex_index(0, new_index)

    def _check_valid_alignment_seq_type(self, seq_type):
        """
        METHOD: _check_valid_alignment_seq_type()

            __STATUS_CHECKER__
            =>Returns BOOLEAN:
                                True if the seq_type is accepted
                                False otherwise
        """
        if not seq_type in BlastHit._VAST:
            raise ValueError("'ref_seq' must be in %s. Recieved 'ref_seq' is %s" % (repr(BlastHit._VAST), seq_type))
        return BlastHit._VAST[seq_type]

    def overlap(self, blastHit):
        # It returns a range between 0<->1 of how much the actual blastHit and the one passed
        # overlap in the query sequence.
        # The overlap is measured OVER THE SHORTEST lenght
        return super(BlastHit, self).overlap(1, 0, blastHit)

    """
        toString
    """
    def _build_slice_alignment(self, sequences, index):
        '''
        This function helps the slicing of the alignment through slice objects as
        an iterable. It is required to support inheritance of the slicing.
        It covers special requirements of child objects when building themselves.

        @param:    sequences
        @pdef:     array of sequences for the new alignment
        @ptype:    {List}

        @param:    index
        @pdef:     transformation and adaptation of the slice object
        @ptype:    {List}

        @return: {IndexedSeqAli}
        '''
        seqInits  = ['' for x in range(self.number_of_sequences)]
        complex_indexes = {}
        for x in range(self.number_of_sequences):
            seqInits[x] = self._sequence_position_from_alignment_position(x, index[0])
            complex_indexes[x] = list(filter(None, self._seq2ali[x]))
            complex_indexes[x] = complex_indexes[x][complex_indexes[x].index(seqInits[x]):]
        patt   = self._alipatt[index[0] - 1:index[1] - 1:index[2]]
        newali = self.__class__(hit = [self._sequenceID, self._length],
                                sequences = [sequences[0], sequences[1], patt],
                                sequence_inits = [seqInits[0], seqInits[1]],
                                iteration      = self._iteration,
                                stats          = [None, None, None, None])
        for refident in complex_indexes:
            newali._update_index(refident, complex_indexes[refident])
            newali._search_segments()
        return newali

    def __str__(self):
        return "\t".join((self.sequenceID, str(self.length), str(self.identities), str(self.positives), str(self.gaps),
                          "{:.3}".format(self.evalue), self.query_seq, self.hit_seq, self.format_positions()))



from .atom     import Atom,           AtomOfAminoAcid,    AtomOfNucleotide

from .residue  import Residue,        ResidueOfAminoAcid, ResidueOfNucleotide

from .chain    import Chain,          ChainOfProtein,     ChainOfNucleotide

from .contacts import PPInterface,    PNInterface,        PHInterface
from .contacts import PPInnerContact, PHInnerContact
from .contacts import Complex,        InnerContacts

from .protein  import Arch

from PDB    import PDB



import BioLib.Structure.PDB
import BioLib.Docking.FTDock

linkerPotential = {'2': (4.274, 11.645), 
                   '3': (5.558, 13.938),
                   '4': (6.651, 15.372),
                   '5': (7.088, 18.662), 
                   '6': (6.631, 22.224),
                   '7': (7.517, 24.180), 
                   '8': (8.312, 25.594), 
                   '9': (9.723, 29.490),
                   '10': (9.666, 33.388),
                   '11': (9.677, 36.293), 
                   '12': (8.461, 38.853), 
                   '13': (10.385, 38.106),
                   '14': (7.937, 40.773),
                   '15': (5.955, 53.153),
                  }

def decoy_linker_energy(static, mobile, ftdock, decoy, nativeMobile = None):
    '''
    Returns 1 if the decoy is compatible with linker statistics, it also returns
    the rmsd with the native solution if nativeMobile is provided
    '''
    staticDecoy = ftdock.get_static_structure(static)
    mobileDecoy = ftdock.get_mobile_structure(static, mobile, decoy)
    linkerDistance = staticDecoy.get_residues()[-1].get_ca_distance(mobileDecoy.get_residues()[0])
    linkerLength = mobileDecoy.get_residues()[0].get_num() - staticDecoy.get_residues()[-1].get_num() -1
    if linkerLength > 15:
         linkerLength = 15
    if linkerPotential[str(linkerLength)][1] > linkerDistance > linkerPotential[str(linkerLength)][0]:
         compatible = 1
    else:
         compatible = 0
    if nativeMobile:
        rmsd = ftdock.get_ligand_RMSD(static, mobile, nativeMobile, decoy)
        print str(decoy.get_id())+'\t'+str(rmsd)+'\t'+str(compatible)
        return compatible, rmsd
    return compatible

def rank_linker_enegy(static, mobile, ftdock, nativeMobile = None):
    '''
    Returns the compatibility of each decoy in a ftdock experiment
    '''
    linkerEnergyDecoy = {}
    for decoy in ftdock.get_decoys():
        linkerEnergyDecoy[decoy.get_id()] = decoy_linker_energy(static, mobile, ftdock, decoy, nativeMobile)
    return linkerEnergyDecoy



import sys
import os
import re


if len(sys.argv)>3:
  fd=open(sys.argv[1],"r")
  ff=open(sys.argv[2],"r")
  outdir=sys.argv[3]
elif len(sys.argv)==3:
  fd=open(sys.argv[1],"r")
  ff=open(sys.argv[2],"r")
  outdir="output.fa"
else:
  print "Run as parse_seq.py 'input_sequences' 'uniprot_database' [FOLDER NAME]"
  print "Missing input list"
  exit(0)

translate={}
for line in ff:
  if line.startswith(">"):
    accession = line.lstrip(">").strip().split("|")[1]
    entry     = line.lstrip(">").strip().split("|")[2].split()[0]
    translate.setdefault(entry,accession)


ff.close()
fo=open(outdir,"w")
for line in fd:
  if line.startswith(">"):
    name_sequence=line.lstrip(">").strip().split("_")
    if len(name_sequence)>2:
     name="_".join(name_sequence[0:-1])
     extension=name_sequence[-1]
    else:
     name="_".join(name_sequence)
     extension=None
    if translate.has_key(name):
     if extension is None: 
       accession=translate[name]
       entry=name
     else:
       accession=translate[name]+"_"+extension
       entry=name+"_"+extension
     fo.write(">sp|%s|%s\n"%(accession,entry))
    else:
     if extension is None:
      entry=name
     else:
      entry=name+"_"+extension
     fo.write(">%s\n"%(entry))
  else:
   fo.write("%s\n"%(line.strip()))

fd.close()
fo.close()
 

    




import os, sys, re
from xml.etree.ElementTree import iterparse

PROTEIN     = 'protein'
INTERACTION = 'interaction'
SIGNATURE   = 'signature'
POSITIVE    = 'positive'
NEGATIVE    = 'negative'
LOOP        = 'loop'
DOMAIN      = 'domain'
MAPPING     = 'mapping'
ALIGNMENT   = 'alignment'
RFRESULT    = 'Random Forest result'
ALL         = 'all'

class ILXMLParser(object): 
    '''
    STRUCTURE OF XML FILE 
    
    ---------------------------------------------------------------------------------------------------------
    Level                                    XMLtags                                 Object/Sub-Object/Buffer
    ---------------------------------------------------------------------------------------------------------

    R: --XML.................................<xml/>..................................ILXMLParser (Buffer)
    0:   +--Protein..........................<protein/>...................................+ILXMLProtein (Object)
         |  +--Name..........................<name/>                                      |     +--name
         |  +--MD5...........................<MD5/>                                       |     +--MD5
         |  +--Sequence......................<sequence/>                                  |     +--Sequence
         |  +--{Alignments}..................<blastSearch><search_type/></blastSearch>    |     |
         |  |   +--AlignmentType.............<ID/>                                        |     |
         |  |   +--AlignmentIteration........<search_iteration/>                          |     |
         |  |   +--..........................<alignments/>                                |     +--{Alignments}
    1:   |  |   +--Alignment.................<align/>.....................................|.....|...+--ILXMLAlignment (Sub-Object)
         |  |      |                                                                      |     |           +--ali_type (<ID>/>)
         |  |      |                                                                      |     |           +--iter_no (<search_iteration/>)
         |  |      +TargetID.................<target_id/>                                 |     |           +--target_id
         |  |      +eValue...................<e-value/>                                   |     |           +--eValue
         |  |      +queryStart...............<query_start/>                               |     |           +--q_start
         |  |      +queryEnd.................<query_end/>                                 |     |           +--q_end
         |  |      +querySequence............<query_seq/>                                 |     |           +--q_seq
         |  |      +hitStart.................<hit_start/>                                 |     |           +--h_start
         |  |      +hitEnd...................<hit_end/>                                   |     |           +--h_end
         |  |      +hitSequence..............<hit_seq/>                                   |     |           +--h_seq
         |  |                                                                             |     |
         |  +--{DomainMappings}..............<domainMapping><alignments/></domainMapping> |     +--{DomainMappings}
         |  |   +--DomainMapping.............<align/>.....................................|.....|...+--ILXMLDomainMapping (Sub-Object)
         |  |      +--targetID...............<target_id/>                                 |     |           +--target_id
         |  |      +--eValue.................<e-value/>                                   |     |           +--eValue
         |  |      +--queryStart.............<query_start/>                               |     |           +--q_start
         |  |      +--queryEnd...............<query_end/>                                 |     |           +--q_end
         |  |      +--hitStart...............<hit_start/>                                 |     |           +--h_start
         |  |      +--hitEnd.................<hit_end/>                                   |     |           +--h_end
         |  |                                                                             |     |
         |  +--{Features}....................<features/>                                  |     +--{Features}
    1:   |      +--Feature...................<feature/>...................................|.........+--ILXMLFeature ([ILXMLLoop, ILXMLDomain]) (Sub-Object)
         |         +--type...................<type/>                                      |                 +--Type ([Loop, Domain])
         |         +--name...................<loop_subClass/>|<SCOP_code/>                |                 +--([subClass, SCOP_code])
         |         +--structureMatch.........<loopMatch/>|<SCOP_match/>                   |                 +--([loopMatch, SCOP_match])
         |         +--alignmentIni...........<align_ini/>                                 |                 +--ali_ini
         |         +--alignmentEnd...........<align_end/>                                 |                 +--ali_end
         |         +--queryFeatureSeq........<query_feature_ali_seq/>                     |                 +--QAliSeq
         |         +--targetFeatureSeq.......<target_feature_ali_seq/>                    |                 +--HAliSeq
         |         +--{domainAssignations}...<SCOP_assignations/>|<domain_mapping/>       |                 +--{mappings} (only ILXMLLoop)
    2:   |             +--domainAssignation..<SCOP_assignation/> |<domain_mapping/>.......|.....................+ILXMLDomainAssignation (Buffer)
         |                +--type............<SCOP_assignation/> |<domain_type/>          |                           +--Type
         |                +--ID..............<SCOP_id/>          |<domain_id/>            |                           +--ID 
         |                +--ini.............<SCOP_ini/>         |<domain_ini/>           |                           +--ini
         |                +--end.............<SCOP_end/>         |<domain_end/>           |                           +--end
         |                                                                                |
    0:   +--Interaction......................<interaction/>...............................+ILXMLInteraction (Object)
            +--Protein1ID....................<P1ID/>                                            +--interactor1_ID (index in ILXMLParser.proteins)
            +--Protein2ID....................<P2ID/>                                            +--interactor2_ID (index in ILXMLParser.proteins)
            +--Splus.........................<Splus/>                                           +--Splus
            +--Sminus........................<Sminus/>                                          +--Sminus
            +--LSR...........................<LSR/>                                             +--LSR
            +--LpVR..........................<LpVR/>                                            +--LpVR
            +--{Random Forest Results}.......<RF_results/>                                      +--{RF_results}
            |   +--Random Forest Result......<RF_result/>.......................................|...+ILXMLInteractionRFResult (Sub-Object)
            |      +--Relative Cost..........<RF_cost/>                                         |         +--cost
            |      +--Prediction.............<RF_prediction/>                                   |         +--prediction <bool>
            |      +--Random Forest Score....<RF_score/>                                        |         +--RF_score
            |      +--{Precisions}...........<inferred_precisions/>                             |         +--{precisions}
            |          +--Precision..........<inferred_precision/>..............................|.............+--ILXMLPrecision (Buffer)
            |             +--Unbalance Ratio.<unbalance_ratio/>                                 |                     +--UR
            |             +--Precision value.<precision/>                                       |                     +--precision
            |             +--Precision error.<error/>                                           |                     +--error
            +--{Interaction Signatures}......<interaction_signatures/>                          +--{signatures}
                +--Interaction Signature.....<interaction_signature/>...............................+ILXMLInteractionSignature (Sub-Object)
                   +--Signature Type.........<IS_type/>                                                   +refMatrix
                   |                                                                                      +elementsType ([LOOP, DOMAIN])
                   +--Protein1 Sign. ID......<P1IS/>                                                      +P1SignID (<tuple>, ref. to self.loops or 
                   |                                                                                      |                           self.domains index)
                   +--Protein2 Sign. ID......<P2IS/>                                                      +P2SignID (<tuple>, ref. to self.loops or 
                   |                                                                                      |                           self.domains index)
                   +--pValue.................<pValue/>                                                    +pValue
    '''


    def __init__(self): 
        self._loops = set()
        self._domains = set()
        self.loops = []
        self.domains = []
        self.proteins = []
        self.interactions = []
        self.active_object = self.reset_active_object()
        self.active_subObj = self.reset_active_subObj()
        self.kwds = self.reset_kwds()
        self.last_output_level = -1
        self.skip_levels = -1

    # SETTERS: CLEAN UP METHODS

    def reset_active_object(self): self.active_object = None
    def reset_active_subObj(self): self.active_subObj = None
    def reset_kwds(self)         : self.kwds          = {}


    # SETTERS ACTIVE OBJECT/SUB-OBJECT CONTROL METHODS

    def set_active_object(self, objType): 
        if   objType == PROTEIN: 
            self.active_object = ILXMLProtein(ILXMLParser=self, ID=len(self.get_proteins()))

        elif objType == INTERACTION: 
            self.active_object = ILXMLInteraction(ILXMLParser=self, ID=len(self.get_interactions()))
            
    def set_active_subObj(self, subObj_generator, parentID, parentType, **kwds): 
        self.active_subObj = subObj_generator(self, parentID, parentType, **kwds)


    # SETTERS: METHODS FOR STORING (REFERENCES TO) OBJECTS/SUB-OBJECTS IN THE PARSER

    def add_protein(self, proteinObj, store=True):
        '''
        o proteinObj: ILXMLProtein object
        o store     : <bool>
        '''
        if store is True: self.proteins.append(proteinObj)
        else            : self.proteins.append(proteinObj.get_name())

    def add_interaction(self, interactionObj, store=True):
        '''
        o proteinObj: ILXMLInteraction object
        o store     : <bool>
        '''
        if store is True: self.interactions.append(interactionObj)
        else            : self.interactions.append( (interactionObj.get_i1name(), interactionObj.get_i2name()) )

    
    def add_feature(self, code, feature_type):
        if   feature_type == LOOP: 
            ref_set  = self._loops
            ref_list = self.get_loops()
        elif feature_type == DOMAIN: 
            ref_set  = self._domains
            ref_list = self.get_domains()

        if not code in ref_set: 
            ref_set.add(code)
            ref_list.append(code)
            
        return ref_list.index(code)


    # PRIVATE SETTERS: METHODS FOR CONTROLLING THE OUTPUT BEHAVIOUR
    def _set_kwds(self, kwds)             : self.kwds              = kwds
    def set_last_output_level(self, level): self.last_output_level = level
    def set_skip_levels(self, level_no)   : self.skip_levels       = level_no


    # GETTERS

    def get_proteins(self)         : return self.proteins
    def get_interactions(self)     : return self.interactions
    def get_loops(self)            : return self.loops
    def get_domains(self)          : return self.domains
    def get_active_object(self)    : return self.active_object
    def get_active_subObj(self)    : return self.active_subObj
    def get_last_output_level(self): return self.last_output_level
    def get_skip_levels(self)      : return self.skip_levels


    # METHODS FOR INITIALIZING OBJECTS/SUB-OBJECTS/BUFFERS
    def _start_protein(self)                   : self.set_active_object( PROTEIN )
    def _start_interaction(self)               : self.set_active_object( INTERACTION )
    def _start_alignment(self, parentProteinID, ali_type, ali_iter): self.set_active_subObj( subObj_generator = ILXMLAlignment, 
                                                                                             parentID         = parentProteinID, 
                                                                                             parentType       = PROTEIN, 
                                                                                             ali_type         = ali_type, 
                                                                                             iter_no          = ali_iter)
    def _start_domain_mapping(self, parentProteinID): self.set_active_subObj( subObj_generator = ILXMLDomainMapping, 
                                                                              parentID         = parentProteinID, 
                                                                              parentType       = PROTEIN)
    def _start_loop_feature(self, parentProteinID): self.set_active_subObj( subObj_generator = ILXMLLoop, 
                                                                            parentID         = parentProteinID, 
                                                                            parentType       = PROTEIN)
    def _start_domain_feature(self, parentProteinID): self.set_active_subObj( subObj_generator = ILXMLDomain, 
                                                                              parentID         = parentProteinID, 
                                                                              parentType       = PROTEIN)

    def _start_RF_result(self, parentInteractionID): self.set_active_subObj( subObj_generator = ILXMLInteractionRFResult, 
                                                                             parentID         = parentInteractionID, 
                                                                             parentType       = INTERACTION)

    def _start_interaction_signature(self, parentInteractionID): self.set_active_subObj( subObj_generator = ILXMLInteractionSignature, 
                                                                                         parentID         = parentInteractionID, 
                                                                                         parentType       = INTERACTION)

    # METHODS FOR ID/HUMAN READABLE CODE CONVERSION
    def get_proteinID(self, protein_code): 
        try: return self.get_proteins().index(protein_code)
        except ValueError:
            protein_codes = [ x.get_name() for x in self.get_proteins() ]
            return protein_codes.index(protein_code)
                              

    def get_featureCode(self, featureID, feature_type): 
        if   feature_type == LOOP  : ref_list = self.get_loops()
        elif feature_type == DOMAIN: ref_list = self.get_domains()
        return ref_list[featureID]

    def get_featureID(self, feature_code, feature_type=None): 
        if feature_type is None: 
            if   feature_code in self._loops()  : feature_type = LOOP
            elif feature_code in self._domains(): feature_type = DOMAIN

        if   feature_type == LOOP  : ref_list = self.get_loops()
        elif feature_type == DOMAIN: ref_list = self.get_domains()
        return ref_list.index(feature_code)
    
    def signatureCode_2_singatureID(self,feature_codes_list):
        if   eval(feature_codes_list)[0] in self._loops  : feature_type = LOOP
        elif eval(feature_codes_list)[0] in self._domains: feature_type = DOMAIN
        else: 
            print self._domains
            print self._loops
            raise ValueError(eval(feature_codes_list)[0])
        
        return [ feature_type, [ self.get_featureID(feature_code, feature_type) for feature_code in eval(feature_codes_list) ]]


    def signatureID_2_signatureCode(self, feature_type, feature_IDs_list): 
        if   feature_type == LOOP  : ref_list = self.get_loops()
        elif feature_type == DOMAIN: ref_list = self.get_domains()

        return [ ref_list[x] for x in feature_IDs_list ]
    

    # OUTPUT CONTROL METHODS
    def _check_output_error(self, object_instance_checkClass, object_instance_str, output_level, output_object=None): 
        '''
        o object_instance_checkClass: class required for the active object or sub-object (depending on output level)
        o object_instance_str       : <str> string describing the class required for the active object or sub-object (depending on output level)
        o output_level              : <int> [0,1]. 0 for output of objects. 1 for output of sub-objects
        o output_object             : object to output
                                      If None (default) the method checks the:
                                      a) self.get_active_object() if output_level is 0
                                      b) self.get_active_subObj() if output_level is 1 
        '''
        if   output_level == 0: 
            obj_str   = 'object'
            check_obj = self.get_active_object()
        elif output_level == 1: 
            obj_str   = 'sub-object'
            check_obj = self.get_active_subObj()

        if   not output_object is None and not isinstance(output_object, object_instance_checkClass): 
            err_msg = 'provided %s must be an instance of %s class(es). Obtained %s is %s\n' %(obj_str, object_instance_str, obj_str, type(check_obj))
            raise ValueError(err_msg)
        elif output_object is None and not isinstance(check_obj, object_instance_checkClass): 
            err_msg = 'active %s must be an instance of %s class(es). Obtained %s is %s\n' %(obj_str, object_instance_str, obj_str, type(check_obj))
            raise ValueError(err_msg)
        elif not output_object is None: report_object = output_object
        else                          : report_object = check_obj

        return report_object

    
    def _control_skip_output(self, output_level):
        if self.get_skip_levels() == output_level: 
            self.set_skip_levels(output_level-1)
            self.set_last_output_level(output_level) # upon return, this level will be skipped, but still need to be annotated as outputed!
            return True

        return False


    def _output_protein_info(self, report_level, protein_object = None, 
                             output_alignments=True, output_domain_mappings=True, output_protein_features=True, output_domain_assignations=True): 
        if self._control_skip_output(output_level = 0) is True: return

        report_object = self._check_output_error(object_instance_checkClass = ILXMLProtein, 
                                                 object_instance_str        = 'ILXMLProtein', 
                                                 output_level               = 0, 
                                                 output_object              = protein_object)

        yield( self.custom_protein_output( report_object, **self.kwds ) )
        if report_level < 1: 
            if output_alignments is True: 
                for alignment in report_object.get_alignments()   : yield( self.custom_alignment_output(alignment, **self.kwds) )

            if output_domain_mappings is True: 
                for domain_mapping in report_object.get_mappings(): yield( self.custom_domain_mapping_output(domain_mapping, **self.kwds) )

            if output_protein_features is True: 
                for domain in report_object.get_domains()         : yield( self.custom_feature_output(domain, DOMAIN, **self.kwds) )
                for loop in report_object.get_loops()             : 
                    yield( self.custom_feature_output(loop, LOOP, **self.kwds) )
                    for domain_assignation in loop.get_mappings(): yield( self.custom_domain_assignation_output(domain_assignation, **self.kwds) )

        self.set_last_output_level(0)
                     

    def _output_protein_alignment_info(self, report_level, alignment_object=None): 
        if self._control_skip_output(output_level = 1) is True: return

        report_object = self._check_output_error(object_instance_checkClass = ILXMLAlignment, 
                                                 object_instance_str        = 'ILXMLAlignment', 
                                                 output_level               = 1, 
                                                 output_object              = alignment_object)

        if self.get_last_output_level() < 1:
            yield( self.custom_protein_output(self.get_active_object(), **self.kwds) )
            self.set_skip_levels(0)
        
        yield( self.custom_alignment_output(report_object, **self.kwds) )
        self.set_last_output_level(1)


    def _output_protein_domain_mapping_info(self, report_level, domain_mapping_object=None, retrieve_results=False):
        if self._control_skip_output(output_level = 1) is True: return

        report_object = self._check_output_error(object_instance_checkClass = ILXMLDomainMapping, 
                                                 object_instance_str        = 'ILXMLDomainMapping', 
                                                 output_level               = 1, 
                                                 output_object              = domain_mapping_object)

        if self.get_last_output_level() < 1: 
            yield( self.custom_protein_output(self.get_active_object(), **self.kwds) )
            self.set_skip_levels(0)

        yield( self.custom_domain_mapping_output(report_object, **self.kwds) )
        self.set_last_output_level(1)


    def _output_protein_features_info(self, report_level, feature_object=None, retrieve_results=False, output_domain_assignations=True):
        if self._control_skip_output(output_level = 1) is True: return

        report_object = self._check_output_error(object_instance_checkClass = ILXMLFeature, 
                                                 object_instance_str        = 'ILXMLLoop or ILXMLDomain', 
                                                 output_level               = 1, 
                                                 output_object              = feature_object)

        if self.get_last_output_level() < 1: 
            yield( self.custom_protein_output(self.get_active_object(), **self.kwds) )
            self.set_skip_levels(0)

        if   isinstance(report_object, ILXMLDomain): yield( self.custom_domain_output(report_object, **self.kwds) )
        elif isinstance(report_object, ILXMLLoop): 
            yield( self.custom_loop_output(report_object, **self.kwds) )
            if report_level < 2: 
                for domain_assignation in report_object.get_mappings(): 
                    yield( self.custom_domain_assignation_output(domain_assignation, **self.kwds) )

        self.set_last_output_level(1)


    def _output_feature_domain_assignation_info(self, report_level, domain_assignation_object, retrieve_results=False): 
        if self._control_skip_output(output_level = 2) is True: return

        if self.get_last_output_level() < 1: 
            yield( self.custom_protein_output(self.get_active_object(), **self.kwds) )
            self.set_skip_levels(0)

        if self.get_last_output_level() < 2: 
            if   isinstance(self.get_active_subObj(), ILXMLLoop)  : yield( self.custom_loop_output(self.get_active_subObj(), **self.kwds) )
            elif isinstance(self.get_active_subOjb(), ILXMLDomain): yield( self.custom_domain_ouptut(self.get_active_subObj(), **self.kwds) )
            self.set_skip_levels(1)
        
        yield( self.custom_RF_precision_output(domain_assignation_object, **self.kwds) )
        self.set_last_output_level(2)


    def _output_interaction_info(self, report_level, interaction_object=None, retrieve_results=False, 
                                 output_interaction_signatures=True, output_RF_results=True, output_RF_precisions=True):     
        if self._control_skip_output(output_level = 0) is True: return

        report_object = self._check_output_error(object_instance_checkClass = ILXMLInteraction, 
                                                 object_instance_str        = 'ILXMLInteraction', 
                                                 output_level               = 0, 
                                                 output_object              = interaction_object)

        yield( self.custom_interaction_output( report_object, **self.kwds ) )

        if report_level < 1:
            if output_interaction_signatures is True: 
                for posSign in report_object.get_positive_signatures(): 
                    yield( self.custom_interaction_signature_output(posSign, POSITIVE, **self.kwds) )
                for negSign in report_object.get_negative_signatures(): 
                    yield( self.custom_interaction_signature_output(negSign, NEGATIVE, **self.kwds) )

            if output_RF_results is True: 
                for RFResult in report_object.get_RFResults(): 
                    yield( self.custom_RF_score_output(RFResult, **self.kwds) )
                    if output_RF_precisions is True: 
                        for precision in RFResult.get_precisions(): yield( self.custom_RF_precision_output(precision, **self.kwds) )

        self.set_last_output_level(0)


    def _output_RF_result_info(self, report_level, RF_result_object=None, retrieve_results=False, output_RF_precisions=True):  
        if self._control_skip_output(output_level = 1) is True: return
    
        report_object = self._check_output_error(object_instance_checkClass = ILXMLInteractionRFResult, 
                                                 object_instance_str        = 'ILXMLInteractionRFResult', 
                                                 output_level               = 1, 
                                                 output_object              = RF_result_object)
        if self.get_last_output_level() < 1: 
            yield( self.custom_interaction_output(self.get_active_object(), **self.kwds) )
            self.set_skip_levels(0)

        yield( self.custom_RF_score_output(report_object, **self.kwds) )

        if report_level < 2: 
            for precision in report_object.get_precisions(): yield( self.custom_RF_precision_output(precision, **self.kwds) )

        self.set_last_output_level(1)


    def _output_RF_precision_info(self, report_level, RF_precision_object, retrieve_results=False): 
        if self._control_skip_output(output_level = 2) is True: return
        
        if self.get_last_output_level() < 1: 
            yield( self.custom_interaction_output(self.get_active_object(), **self.kwds) )
            self.set_skip_levels(0)

        if self.get_last_output_level() < 2: 
            yield( self.custom_RF_score_output(self.get_active_subObj(), **self.kwds) )
            self.set_skip_levels(1)

        yield( self.custom_RF_precision_output(RF_precision_object, **self.kwds) )
        self.set_last_output_level(2)


    def _output_interaction_signature_info(self, report_level, interaction_signature_object=None, retrieve_results=False): 
        if self._control_skip_output(output_level = 1) is True: return
 
        report_object = self._check_output_error(object_instance_checkClass = ILXMLInteractionSignature, 
                                                 object_instance_str        = 'ILXMLInteractionSignature', 
                                                 output_level               = 1, 
                                                 output_object              = interaction_signature_object)

        if self.get_last_output_level() < 1: 
            yield( self.custom_interaction_output(self.get_active_object(), **self.kwds) )
            self.set_skip_levels(0)
        
        yield( self.custom_interaction_signature_output(report_object, report_object.get_refMatrix(), **self.kwds) )
        self.set_last_output_level(1)


    # CUSTOM OUTPUT METHODS (should be overwritten by user for output style modification) 
    def custom_protein_output(self, protein_object, **kwds)                                    : return repr(protein_object)    
    def custom_alignment_output(self, alignment_subObj, **kwds)                                : return repr(alignment_subObj)
    def custom_domain_mapping_output(self, domain_mapping_subObj, **kwds)                      : return repr(domain_mapping_subObj)
    def custom_feature_output(self, feature_subObj, feature_type, **kwds)                      : return repr(feature_subObj)
    def custom_domain_assignation_output(self, domain_assign_buffer, **kwds)                   : return repr(domain_assign_buffer)
    def custom_interaction_output(self, interaction_object, **kwds)                            : return repr(interaction_object)
    def custom_interaction_signature_output(self, int_signature_subObj, signature_type, **kwds): return repr(int_signature_subObj)
    def custom_RF_score_output(self, RF_score_subObj, **kwds)                                  : return repr(RF_score_subObj)
    def custom_RF_precision_output(self, RF_precision_buffer, **kwds)                          : return repr(RF_precision_buffer)

    def custom_domain_output(self, domain_subObj, **kwds): return self.custom_feature_output(domain_subObj, DOMAIN, **kwds)
    def custom_loop_output(self, loop_subObj, **kwds)    : return self.custom_feature_output(loop_subObj, LOOP, **kwds)


    # PARSER METHODS
    def results_parser(self, xml_file, report_level=0, skip_alignment_type=[], output_proteins=True, 
                       output_alignments=True, output_domain_mappings=True, output_protein_features=True, output_domain_assignations=True, 
                       output_interactions=True, output_interaction_signatures=True, output_RF_results=True, output_RF_precisions=True, **kwds):
        
        '''
        o xml_file                     : <str>  iLoops XML results file to be parsed
        o report_level                 : <int>  contained in set([-1,0,1,2])
                                         If -1: Only report after parsing the full XML file. Stores all information of the XML document into memmory. 
                                         If  0: Only report after reading a complete protein or interaction. 
                                         If  1: Report after reading a level 1 object 
                                                (Alignment, DomainMapping, Feature for proteins)
                                                (InteractionSignature, RFResult for interactions)
                                         If  2: Report after reading a level 2 object
        o skip_alignment_type          : <list> containing all the types of alignments to be skiped during parse. 
                                         Use [ALL,] to skip all (do not parse alignments information)
        o output_proteins              : <bool> If True, info about proteins is outputed through self.custom_protein_output() method
        o output_alignments            : <bool> If True, info about alignments is outputed through self.custom_alignment_output() method
        o output_domain_mappings       : <bool> If True, info about domain mappings is outputed through self.custom_domain_mapping_output() method
                                                Domain mappins are domains mapped to the whole protein. 
        o output_protein_features      : <bool> If True, info about features (loops, domains) is outputed through:
                                                         Domains: self.custom_loop_output()
                                                         Loops  : self.custom_domain_output()
                                                A generic self.custom_fature_output() is also available. 
        o output_domain_assignations   : <bool> If True, info about domain assignations is outputed through self.custom_domain_assignation_output() method. 
                                                Domain assignations are domains assigned (mapped) to loop features
        o output_interactions          : <bool> If True, info about interactions is outputed through self.custom_interaction_output() method
        o output_interaction_signatures: <bool> If True, info about int. signatures is outputed through self.custom_interaction_signature_output() method.
        o output_RF_results            : <bool> If True, info about random forest results is outputed trhough self.custom_RF_score_output() method
        o output_RF_precisions         : <bool> If True, info about unabalance ratios and inferred precisions of the random forest results is outputed
                                                         through self.custom_RF_precision_output method()
        '''

        # SET STORE CONDITIONS
        store_proteins               = False
        store_interactions           = False
        store_alignments             = True
        store_domain_mappings        = True
        store_protein_features       = True
        store_domain_assignations    = True
        store_interaction_signatures = True
        store_RF_results             = True
        store_RF_precisions          = True

        # CONROL OUTPUT BEHAVIOUR
        self._set_kwds(kwds)
        parsed_interactions              = False

        if report_level > 1: 
            store_domain_assignations    = False
            store_RF_precisions          = False
        if report_level > 0: 
            store_alignments             = False
            store_domain_mappings        = False
            store_protein_features       = False
            store_interaction_signatures = False
            store_RF_results             = False
        if report_level == -1: 
            store_proteins               = True
            store_interactions           = True

            
        # CREATE ITERATOR FOR XML PARSER
        context = iterparse(xml_file, ("start", "end"))
        context = iter(context)
        event, root = context.next() #drop root, usually '<xml>'

        # ITERATE OVER ELEMENTS IN XML FILE
        for event, element in context: 
            if event == "start": 
                #if element.tag == "feature" or element.tag == "type": sys.stderr.write("<START> %s: %s\n" %(element.tag, element.text)) 

                # LEVEL 0: START PROTEIN 
                if   element.tag == "protein"           : self._start_protein()

                # LEVEL 1 (PROTEIN): START ALIGNMENT INFORMATION (ALIGNMENTS AND DOMAIN MAPPINGS)
                elif element.tag == "blastSearch"       : current_ali_starter = self._start_alignment
                elif element.tag == "domainMapping"     : current_ali_starter = self._start_domain_mapping
                elif element.tag == "align":
                    if   current_ali_starter == self._start_alignment:
                        if not (element.text or ALL) in skip_alignment_type: current_ali_starter(parentProteinID = self.get_active_object().get_ID(),
                                                                                                 ali_type        = current_ali_type,
                                                                                                 ali_iter        = current_ali_iter)
                    elif current_ali_starter == self._start_domain_mapping: current_ali_starter(parentProteinID = self.get_active_object().get_ID())

                # LEVEL 1 (PROTEIN): START FEATURES 
                # Since it is required to distinguish between loops and domains, this start is done on an 'end' event (element.tag = "type")

                # LEVEL 0: START INTERACTION
                elif element.tag == "interaction": 
                    self._start_interaction()
                    if parsed_interactions is False: 
                        self.set_last_output_level(-1)
                        parsed_interactions = True
                    
                # LEVEL 1 (INTERACTION): START RANDOM FOREST SCORES
                elif element.tag == "RF_result":  self._start_RF_result(self.get_active_object().get_ID())

                # LEVEL 1 (INTERACTION): START INTERACTION SIGNATURES
                elif element.tag == "interaction_signature": self._start_interaction_signature(self.get_active_object().get_ID())
                

            if event == "end":
                #if element.tag == "feature" or element.tag == "type": sys.stderr.write("<END> %s\n" %element.tag)

                # PROTEIN LEVEL O: PROTEIN INFORMATION
                if   element.tag == "name"    : self.get_active_object().set_name(element.text)
                elif element.tag == "MD5"     : self.get_active_object().set_MD5(element.text)
                elif element.tag == "sequence": self.get_active_object().set_sequence(element.text)
                

                # PROTEIN LEVEL 1: ALIGNMENTS AND DOMAIN MAPPINGS
                elif element.tag == "ID"                : current_ali_type    = element.text
                elif element.tag == "search_iteration"  : current_ali_iter    = element.text
                elif element.tag== "target_id" and ( isinstance(self.get_active_subObj(), ILXMLAlignment) or \
                                                     isinstance(self.get_active_subObj(), ILXMLDomainMapping) ):
                    self.get_active_subObj().set_targetID(element.text)
                elif element.tag== "e-value" and ( isinstance(self.get_active_subObj(), ILXMLAlignment) or \
                                                   isinstance(self.get_active_subObj(), ILXMLDomainMapping) ):
                    self.get_active_subObj().set_eValue(element.text)
                elif element.tag== "query_start" and ( isinstance(self.get_active_subObj(), ILXMLAlignment) or \
                                                       isinstance(self.get_active_subObj(), ILXMLDomainMapping) ):
                    self.get_active_subObj().set_QStart(element.text)
                elif element.tag== "query_end" and ( isinstance(self.get_active_subObj(), ILXMLAlignment) or \
                                                     isinstance(self.get_active_subObj(), ILXMLDomainMapping) ):
                    self.get_active_subObj().set_QEnd(element.text)
                elif element.tag== "query_seq" and isinstance(self.get_active_subObj(), ILXMLAlignment):
                    self.get_active_subObj().set_QSeq(element.text)
                elif element.tag== "hit_start" and ( isinstance(self.get_active_subObj(), ILXMLAlignment) or \
                                                     isinstance(self.get_active_subObj(), ILXMLDomainMapping) ):
                    self.get_active_subObj().set_HStart(element.text)
                elif element.tag== "hit_end" and ( isinstance(self.get_active_subObj(), ILXMLAlignment) or \
                                                   isinstance(self.get_active_subObj(), ILXMLDomainMapping) ):
                    self.get_active_subObj().set_HEnd(element.text)
                elif element.tag== "hit_seq" and isinstance(self.get_active_subObj(), ILXMLAlignment):
                    self.get_active_subObj().set_HSeq(element.text)


                # PROTEIN LEVEL 1: CLOSING ALIGNMENTS AND DOMAIN MAPPINGS
                elif element.tag == "align"  : 
                    if isinstance(self.get_active_object(), ILXMLProtein): 
                        if isinstance(self.get_active_subObj(), ILXMLAlignment): 
                            if not (element.text or ALL) in skip_alignment_type and store_alignments is True: 
                                self.get_active_object().add_alignment(self.get_active_subObj())
                            if output_alignments is True and report_level >= 1: 
                                for x in self._output_protein_alignment_info(report_level = report_level): 
                                    if not x is None: yield x

                        elif isinstance(self.get_active_subObj(), ILXMLDomainMapping): 
                            if store_domain_mappings is True: self.get_active_object().add_mapping(self.get_active_subObj())
                            if output_domain_mappings is True and report_level >= 1: 
                                for x in self._output_protein_domain_mappings(report_level = report_level): 
                                    if not x is None: yield x
                    self.reset_active_subObj()

                elif element.tag == "search_type": 
                    current_ali_iter = None
                    current_ali_type = None

                elif element.tag == "blastSearch" or element.tag == "domainMapping": current_ali_starter = None
                
 
                # PROTEIN LEVEL 1: STARTING FEATURES 
                #                  !!!!SPECIAL: To ensure feature type (loop or domain) the text of the element should be read!
                #                               For this reason the starting of features is done on an 'end' event. 
                elif isinstance(self.get_active_object(), ILXMLProtein) and element.tag == "type": 
                    if   element.text == "loop"  : self._start_loop_feature(parentProteinID = self.get_active_object().get_ID())
                    elif element.text == "domain": self._start_domain_feature(parentProteinID = self.get_active_object().get_ID())    


                # PROTEIN LEVEL 1: FEATURES (feature specific)
                elif element.tag == "loop_subClass"         : self.get_active_subObj().set_subClass(element.text)
                elif element.tag == "SCOP_code"             : self.get_active_subObj().set_SCOP_code(element.text)
                elif element.tag == "loop_match"            : self.get_active_subObj().set_loopMatch(element.text)
                elif element.tag == "SCOP_match"            : self.get_active_subObj().set_SCOP_match(element.text)

                # PROTEIN LEVEL 1: FEATURES (common)
                elif element.tag == "align_ini"             : self.get_active_subObj().set_ali_ini(element.text)
                elif element.tag == "align_end"             : self.get_active_subObj().set_ali_end(element.text)
                elif element.tag == "query_feature_ali_seq" : self.get_active_subObj().set_QAliSeq(element.text)
                elif element.tag == "target_feature_ali_seq": self.get_active_subObj().set_HAliSeq(element.text)


                # PROTEIN LEVEL 2: DOMAIN ASSIGNATIONS
                elif element.tag == "SCOP_id" or element.tag == "domain_id"  : current_assignation_id   = element.text
                elif element.tag == "SCOP_ini" or element.tag == "domain_ini": current_assignation_ini  = element.text
                elif element.tag == "SCOP_end" or element.tag == "domain_end": current_assignation_end  = element.text


                # PROTEIN LEVEL 2: CLOSING DOMAIN ASSIGNATIONS
                elif element.tag == "SCOP_assignation" or element.tag == "domain_type": 
                    if   element.tag == "SCOP_assignation": current_assignation_type = 'SCOP'
                    elif element.tag == "domain_type"     : current_assignation_type = element.text

                    this_assignation = ILXMLDomainAssignation(current_assignation_type, current_assignation_id, 
                                                              current_assignation_ini, current_assignation_end)
                    if isinstance(self.get_active_subObj(), ILXMLFeature) and isinstance(self.get_active_object(), ILXMLProtein):
                        if store_domain_assignations is True: self.get_active_subObj().add_mapping(this_assignation)
                        if output_domain_assignations is True and report_level >= 2: 
                            for x in self._output_feature_domain_assignation_info(report_level              = report_level, 
                                                                                  domain_assignation_object = this_assignation): 
                                if not x is None: yield x
                    current_assignation_type = None
                    current_assignation_id   = None
                    current_assignation_ini  = None
                    current_assignation_end  = None
                                        

                # PROTEIN LEVEL 1: CLOSING FEATURES
                elif element.tag == 'feature': 
                    if isinstance(self.get_active_subObj(), ILXMLFeature) and isinstance(self.get_active_object(), ILXMLProtein):                       
                        self.add_feature(self.get_active_subObj().get_code(), self.get_active_subObj().get_type())
                        if store_protein_features is True: self.get_active_object().add_feature(self.get_active_subObj()) 
                        if output_protein_features is True and report_level >= 1: 
                            for x in self._output_protein_features_info(report_level               = report_level,
                                                                        output_domain_assignations = output_domain_assignations): 
                                if not x is None: yield x
                    self.reset_active_subObj()


                # LEVEL 0: CLOSING PROTEIN
                elif element.tag == 'protein': 
                    self.add_protein(self.get_active_object(), store=store_proteins)
                    if output_proteins is True and report_level >= 0: 
                        for x in self._output_protein_info(report_level               = report_level, 
                                                           output_alignments          = output_alignments, 
                                                           output_domain_mappings     = output_domain_mappings,
                                                           output_protein_features    = output_protein_features, 
                                                           output_domain_assignations = output_domain_assignations): 
                            if not x is None: yield x
                    self.reset_active_object()
                           

                # INTERACTION LEVEL 0: INTERACTION INFO
                elif element.tag == 'P1ID': self.get_active_object().set_interactor1_ID(self.get_proteinID(element.text))
                elif element.tag == 'P2ID': self.get_active_object().set_interactor2_ID(self.get_proteinID(element.text))
                elif element.tag == 'Splus' : self.get_active_object().set_Splus(element.text)
                elif element.tag == 'Sminus': self.get_active_object().set_Sminus(element.text)
                elif element.tag == 'LSR'   : self.get_active_object().set_LSR(element.text)
                elif element.tag == 'LpVR'  : self.get_active_object().set_LpVR(element.text)


                # INTERACTION LEVEL 1: RANDOM FOREST INFO
                elif element.tag == "RF_cost"                                : self.get_active_subObj().set_cost(element.text)
                elif element.tag == "RF_prediction" and element.text == "YES": self.get_active_subObj().set_prediction(True)
                elif element.tag == "RF_prediction" and element.text == "NO" : self.get_active_subObj().set_prediction(False)
                elif element.tag == "RF_score"                               : self.get_active_subObj().set_RFscore(element.text)


                # INTERACTION LEVEL 2: UNBALANCE RATIOS AND PRECISIONS INFO
                elif element.tag == "unbalance_ratio"                        : current_UR            = element.text
                elif element.tag == "precision"                              : current_precision     = element.text
                elif element.tag == "error"                                  : current_error         = element.text


                # INTERACTION LEVEL 2: CLOSING UNBALANCE RATIONS AND PRECISIONS
                elif element.tag == "inferred_precision": 
                    if isinstance(self.get_active_object(), ILXMLInteraction):
                        this_precision = ILXMLPrecision(current_UR, current_precision, current_error)
                        if store_RF_precisions is True: self.get_active_subObj().add_precision(this_precision)
                        if output_RF_precisions is True and report_level >=2: 
                            for x in self._output_RF_precision_info(report_level        = report_level, 
                                                                    RF_precision_object = this_precision): 
                                if not x is None: yield x
                    current_UR        = None
                    current_precision = None
                    current_error     = None


                # INTERACTION LEVEL 1: CLOSING RANDOM FOREST SCORES
                elif element.tag == "RF_result": 
                    if isinstance(self.get_active_object(), ILXMLInteraction) and isinstance(self.get_active_subObj(), ILXMLInteractionRFResult):
                        if store_RF_results is True: self.get_active_object().add_RFResult(self.get_active_subObj())
                        if output_RF_results is True and report_level >=1: 
                            for x in self._output_RF_result_info(report_level         = report_level, 
                                                                 output_RF_precisions = output_RF_precisions): 
                                if not x is None: yield x
                    self.reset_active_subObj()


                # INTERACTION LEVEL 1: INTERACTION SIGNATURES INFO
                elif element.tag == "IS_type"              : self.get_active_subObj().set_refMatrix(element.text)
                elif element.tag == "pValue"               : self.get_active_subObj().set_pValue(element.text)
                elif element.tag == "P1IS"                 : 
                    signature_type, signatureIDs = self.signatureCode_2_singatureID(element.text)
                    self.get_active_subObj().set_elementsType(signature_type)
                    self.get_active_subObj().set_P1SignID(tuple(signatureIDs))
                elif element.tag == "P2IS"                 : self.get_active_subObj().set_P2SignID(tuple(self.signatureCode_2_singatureID(element.text)[1]))


                # INTERACTION LEVEL 1: CLOSING INTERACTION SIGNATURES
                elif element.tag == "interaction_signature": 
                    if isinstance(self.get_active_object(), ILXMLInteraction) and isinstance(self.get_active_subObj(), ILXMLInteractionSignature):
                        if store_interaction_signatures is True: self.get_active_object().add_signature(self.get_active_subObj())
                        if output_interaction_signatures is True and report_level >=1: 
                            for x in self._output_interaction_signature_info(report_level = report_level): 
                                if not x is None: yield x
                    self.reset_active_subObj()


                # LEVEL 0: CLOSING INTERACTION
                elif element.tag == "interaction" and isinstance(self.get_active_object(), ILXMLInteraction):
                    self.add_interaction(self.get_active_object(), store=store_interactions)
                    if output_interactions is True and report_level >= 0: 
                        for x in self._output_interaction_info(report_level                  = report_level, 
                                                               output_interaction_signatures = output_interaction_signatures, 
                                                               output_RF_results             = output_RF_results, 
                                                               output_RF_precisions          = output_RF_precisions): 
                            if not x is None: yield x
                    self.reset_active_object()
                                        

                # CLEAN
                element.clear()

        # CLEAN
        root.clear()

                
        # OUTPUT COMPLETE DOCUMENT IF report_level == -1
        if report_level == -1: 
            if output_proteins is True:
                for protein in self.get_proteins(): self._output_protein_info(report_level               = report_level, 
                                                                              protein_object             = protein,
                                                                              output_alignments          = output_alignments, 
                                                                              output_domain_mappings     = output_domain_mappings,
                                                                              output_protein_features    = output_protein_features, 
                                                                              output_domain_assignations = output_domain_assignations)

            if output_interactions is True:
                for interaction in self.get_interactions(): self._output_interaction_info(report_level                  = report_level, 
                                                                                          interaction_object            = interaction,
                                                                                          output_interaction_signatures = output_interaction_signatures, 
                                                                                          output_RF_results             = output_RF_results, 
                                                                                          output_RF_precisions          = output_RF_precisions)

            yield( self.get_proteins() )
            yield( self.get_interactions() )


    def errors_parser(self, xml_file, report_level=0, **kwds):
        # CONTROL OUTPUT BEHAVIOUR
        #self._set_kwds(kwds)
        raise NotImplementedError("Not Implemented yet!\n")

    
    def simple_parse_results(self, xml_file, out_fd=sys.stdout, report_level=0, **kwds):
        # PARSER
        for result_item in self.results_parser(xml_file=xml_file, report_level=report_level, **kwds): 
            # process the customized result item
            if report_level > -1: out_fd.write(result_item)
            else: 
                for i in result_item: out_fd.write(repr(i))


    def parse_results(self, xml_file, out_fd=sys.stdout, report_level=0, **kwds):
        self.simple_parse_results(xml_file=xml_file, out_fd=out_fd, report_level=report_level, **kwds)


    def parse_errors(self, xml_file, out_fd=sys.stdout, report_level=0, **kwds):
        #for result_item in self.erorrs_parser(xml_file=xml_file, report_level=report_level, **kwds): 
        #    # do something interseting with the customized result item
        #    pass 
        raise NotImplementedError("Not Implemented yet!\n")
    


class ILXMLObject(object): 
    def __init__(self, ILXMLParser, ID=None, Type=None):
        self.parser = ILXMLParser
        self.ID     = ID
        self.Type   = Type

    # SETTERS
    def _set_parser(self, parser): self.parser=parser

    # GETTERS
    def get_ID(self)    : return self.ID
    def get_type(self)  : return self.Type
    def get_parser(self): return self.parser

    # OUTPUT
    def __repr__(self): raise NotImplementedError("Not implemented yet!\n") # overwritten by child classes


class ILXMLSubObj(ILXMLObject): 
    def __init__(self, ILXMLParser, parentID, parentType, ID=None, Type=None):
        super(ILXMLSubObj, self).__init__(ILXMLParser=ILXMLParser, ID=ID, Type=Type)
        self.parentID   = parentID
        self.parentType = parentType

    # GETTERS
    def get_parentID(self)  : return self.parentID
    def get_parentType(self): return self.parentType

    # OUTPUT
    def __repr__(self): raise NotImplementedError("Not implemented yet!\n") # overwritten by child classes
    

class ILXMLFeature(ILXMLSubObj):
    def __init__(self, ILXMLParser, ID=None, Type=None, parentID=None, parentType=None): 
        super(ILXMLFeature, self).__init__(ILXMLParser, parentID, parentType, ID, Type)
        self.ali_ini = None # overwritten by child classes
        self.ali_end = None # overwritten by child classes
        self.QAliSeq = None # overwritten by child classes
        self.HAliSeq = None # overwritten by child classes

    # SETTERS
    def set_ali_ini(self, ali_ini): self.ali_ini = ali_ini
    def set_ali_end(self, ali_end): self.ali_end = ali_end
    def set_QAliSeq(self, QAliSeq): self.QAliSeq = QAliSeq
    def set_HAliSeq(self, HAliSeq): self.HAliSeq = HAliSeq

    # GETTERS
    def get_ali_ini(self): return self.ali_ini
    def get_ali_end(self): return self.ali_end
    def get_QAliSeq(self): return self.QAliSeq
    def get_HAliSeq(self): return self.HAliSeq
    def get_code(self)    : raise NotImplementedError("Not implemented yet!\n") # overwritten by child classes

    # OUTPUT
    def __repr__(self): raise NotImplementedError("Not implemented yet!\n") # overwritten by child classes


class ILXMLProtein(ILXMLObject):

    def __init__(self, ILXMLParser, ID, name=None, MD5=None, sequence=None, appendAlignments=True, appendMappings=True, appendFeatures=True):
        super(ILXMLProtein, self).__init__(ILXMLParser, ID, PROTEIN)

        self.name     = name
        self.MD5      = MD5
        self.sequence = sequence
        
        self.appendAlignments = appendAlignments
        self.appendMappings   = appendMappings
        self.appendFeatures   = appendFeatures

        self.reset_alignments()
        self.reset_mappings()
        self.reset_features()

    # SETTERS (information attributes)
    def set_name(self, name)        : self.name     = name
    def set_MD5(self, MD5)          : self.MD5      = MD5
    def set_sequence(self, sequence): self.sequence = sequence

    # SETTERS (container attributes)
    def add_alignment(self, alignmentSubObj): 
        if self.get_appendAlignments() is True: self.alignments.append(alignmentSubObj)
    def add_mapping(self, mappingSubObj): 
        if self.get_appendMappings() is True: self.mappings.append(mappingSubObj)
    def add_feature(self, featureSubObj): 
        if self.get_appendFeatures() is True: self.features.append(featureSubObj)

    # SETTERS (container attributes clean up)
    def reset_alignments(self): self.alignments = []
    def reset_mappings(self)  : self.mappings   = []
    def reset_features(self)  : self.features   = []

    # GETTERS (information attributes)
    def get_name(self)    : return self.name
    def get_MD5(self)     : return self.MD5
    def get_sequence(self): return self.sequence

    # GETTERS (containter controller attributes)
    def get_appendAlignments(self): return self.appendAlignments
    def get_appendMappings(self)  : return self.appendMappings
    def get_appendFeatures(self)  : return self.appendFeatures

    # GETTERS (containers)
    def get_alignments(self): return self.alignments
    def get_mappings(self)  : return self.mappings
    def get_features(self)  : return self.features
    def get_loops(self)     : return [ x for x in self.get_features() if isinstance(x, ILXMLLoop) is True ]
    def get_domains(self)   : return [ x for x in self.get_features() if isinstance(x, ILXMLDomain) is True ]

    # OUTPUT
    def __repr__(self): 
        return "item = PROTEIN, Protein = %s, MD5 = %s, sequence = %s\n" %(self.get_name(), self.get_MD5(), self.get_sequence())


class ILXMLAlignment(ILXMLSubObj): 

    def __init__(self, ILXMLParser, parentID, parentType, 
                 ali_type, iter_no, target_id=None, eValue=None, q_start=None, q_end=None, q_seq=None, h_start=None, h_end=None, h_seq=None): 

        super(ILXMLAlignment, self).__init__(ILXMLParser = ILXMLParser, 
                                             ID          = None, 
                                             Type        = ALIGNMENT,
                                             parentID    = parentID, 
                                             parentType  = parentType)

        self.ali_type  = ali_type
        self.iter_no   = iter_no
        self.target_id = target_id
        self.eValue    = eValue
        self.q_start   = q_start
        self.q_end     = q_end
        self.q_seq     = q_seq
        self.h_start   = h_start
        self.h_end     = h_end
        self.h_seq     = h_seq

    # SETTERS
    def set_targetID(self, targetID): self.targetID = targetID
    def set_eValue(self, eValue)    : self.eValue   = eValue
    def set_QStart(self, QStart)    : self.q_start   = QStart
    def set_QEnd(self, QEnd)        : self.q_end     = QEnd
    def set_QSeq(self, QSeq)        : self.q_seq     = QSeq
    def set_HStart(self, HStart)    : self.h_start   = HStart
    def set_HEnd(self, HEnd)        : self.h_end     = HEnd
    def set_HSeq(self, HSeq)        : self.h_seq     = HSeq

    # GETTERS
    def get_aliType(self) : return self.ali_type
    def get_iterNo(self)  : return self.iter_no
    def get_targetID(self): return self.targetID
    def get_eValue(self)  : return self.eValue
    def get_QStart(self)  : return self.q_start
    def get_QEnd(self)    : return self.q_end
    def get_QSeq(self)    : return self.q_seq
    def get_HStart(self)  : return self.h_start
    def get_HEnd(self)    : return self.h_end
    def get_HSeq(self)    : return self.h_seq

    # OUTPUT
    def __repr__(self): 
        ret_str  = "\titem = ALIGNMENT, Alignment type = %s, iteration = %s, target ID = %s, " %(self.get_aliType(), self.get_iterNo(), self.get_targetID())
        ret_str += "eValue =%s, query start = %s, query end = %s, " %(self.get_eValue(), self.get_QStart(), self.get_QEnd())
        ret_str += "hit start = %s, hit end = %s\n" %(self.get_HStart(), self.get_HEnd())
        ret_str += "\t**ALIGNMENT:\n\t**Query: %s\n\t**Hit  : %s\n" %(self.get_QSeq(), self.get_HSeq())
        return ret_str


class ILXMLDomainMapping(ILXMLSubObj): 
    def __init__(self, ILXMLParser, parentID, parentType, targetID=None, eValue=None, q_start=None, q_end=None, h_start=None, h_end=None): 
        super(ILXMLDomainMapping, self).__init__(ILXMLParser = ILXMLParser, 
                                                 ID          = None, 
                                                 Type        = MAPPING, 
                                                 parentID    = parentID, 
                                                 parentType  = parentType) # PROTEIN

        self.targetID = targetID
        self.eValue   = eValue
        self.q_start  = q_start
        self.q_end    = q_end
        self.h_start  = h_start
        self.h_end    = h_end

    # SETTERS
    def set_targetID(self, targetID): self.targetID = targetID
    def set_eValue(self, eValue)    : self.eValue   = eValue
    def set_QStart(self, QStart)    : self.q_start  = QStart
    def set_QEnd(self, QEnd)        : self.q_end    = QEnd
    def set_HStart(self, HStart)    : self.h_start  = HStart
    def set_HEnd(self, HEnd)        : self.h_end    = HEnd

    # GETTERS
    def get_targetID(self): return self.targetID
    def get_eValue(self)    : return self.eValue
    def get_QStart(self)    : return self.q_start
    def get_QEnd(self)      : return self.q_end
    def get_HStart(self)    : return self.h_start
    def get_HEnd(self)      : return self.h_end

    # OUTPUT
    def __repr__(self): 
        return "\titem = DOMAIN MAPPING, Target ID = %s, eValue=%s, query start = %s, query end = %s, hit start = %s, hit end = %s\n" %(self.get_targetID(),
                                                                                                                                        self.get_eValue(), 
                                                                                                                                        self.get_QStart(), 
                                                                                                                                        self.get_QEnd(), 
                                                                                                                                        self.get_HStart(), 
                                                                                                                                        self.get_HEnd())


class ILXMLLoop(ILXMLFeature):
    def __init__(self, ILXMLParser, parentID, parentType, subClass=None, loopMatch=None, ali_ini=None, ali_end=None, QAliSeq=None, HAliSeq=None):
        super(ILXMLLoop, self).__init__(ILXMLParser = ILXMLParser, 
                                        ID          = None,
                                        Type        = LOOP, 
                                        parentID    = parentID, 
                                        parentType  = parentType) # PROTEIN
        self.subClass  = subClass
        self.loopMatch = loopMatch
        self.ali_ini   = ali_ini
        self.ali_end   = ali_end
        self.QAliSeq   = QAliSeq
        self.HAliSeq   = HAliSeq
        self.mappings  = []

    # SETTERS (information attributes)
    def set_subClass(self, subClass)    : self.subClass  = subClass
    def set_loopMatch(self, loopMatch)  : self.loopMatch = loopMatch

    # SETTERS (containers)
    def add_mapping(self, mappingObject): self.mappings.append(mappingObject)

    # GETTERS (overwrite parent inheritance)
    def get_code(self)     : return self.get_subClass()

    # GETTERS 
    def get_subClass(self) : return self.subClass
    def get_loopMatch(self): return self.loopMatch
    def get_mappings(self) : return self.mappings

    # OUTPUT
    def __repr__(self): 
        itemID = "item = LOOP"
        return "\t%s, ArchDB code = %s, ArhcDB match = %s, ini = %s, end =%s\n\t**ALIGNMENT:\n\t**Query: %s\n\t**Hit  : %s\n" %(itemID, 
                                                                                                                                self.get_subClass(), 
                                                                                                                                self.get_loopMatch(), 
                                                                                                                                self.get_ali_ini(), 
                                                                                                                                self.get_ali_end(), 
                                                                                                                                self.get_QAliSeq(), 
                                                                                                                                self.get_HAliSeq())

class ILXMLDomain(ILXMLFeature):
    def __init__(self, ILXMLParser, parentID, parentType, SCOP_code=None, SCOP_match=None, ali_ini=None, ali_end=None, QAliSeq=None, HAliSeq=None):
        super(ILXMLDomain, self).__init__(ILXMLParser = ILXMLParser, 
                                          ID          = None, 
                                          Type        = DOMAIN, 
                                          parentID    = parentID, 
                                          parentType  = parentType) # PROTEIN
        self.SCOP_code  = SCOP_code
        self.SCOP_match = SCOP_match
        self.ali_ini    = ali_ini
        self.ali_end    = ali_end
        self.QAliSeq    = QAliSeq
        self.HAliSeq    = HAliSeq

    # SETTERS
    def set_SCOP_code(self, SCOP_code)  : self.SCOP_code  = SCOP_code
    def set_SCOP_match(self, SCOP_match): self.SCOP_match = SCOP_match 
    
    # GETTERS (overwrite parent inheritance)
    def get_code(self): return self.get_SCOP_match()

    # GETTERS
    def get_SCOP_code(self) : return self.SCOP_code
    def get_SCOP_match(self): return self.SCOP_match

    # OUTPUT
    def __repr__(self): 
        itemID = 'item = DOMAIN'
        return "\t%s, SCOP code = %s, SCOP match = %s, ini = %s, end =%s\n\t**ALIGNMENT:\n\t**Query: %s\n\t**Hit  : %s\n" %(itemID, 
                                                                                                                            self.get_SCOP_code(), 
                                                                                                                            self.get_SCOP_match(), 
                                                                                                                            self.get_ali_ini(), 
                                                                                                                            self.get_ali_end(), 
                                                                                                                            self.get_QAliSeq(), 
                                                                                                                            self.get_HAliSeq())

class ILXMLDomainAssignation(object):
    def __init__(self, Type, ID, ini, end):
        self.Type = Type
        self.ID   = ID
        self.ini  = ini
        self.end  = end 

    # SETTERS
    # Not reequired. A "buffer" object is initialized with all its contents. 

    # GETTERS
    def get_type(self): return self.Type
    def get_ID(self)  : return self.ID
    def get_ini(self) : return self.ini
    def get_end(self) : return self.end

    # OUTPUT
    def __repr__(self):
        return "\t\titem = DOMAIN ASSIGNATION, Type = %s, Domain code = %s, ini = %s, end = %s\n" %(self.get_type(), self.get_ID(), 
                                                                                                    self.get_ini(), self.get_end())


class ILXMLProteinSignature(ILXMLSubObj):
    def __init__(self, ILXMLParser, parentID, parentType, ID, signature_type, participantsIDs):
        '''
        o ID             : <tuple> tuple containing the IDs (as per ILXMLParser object) of the features (elements) in the signature
        o parentID       : <int>   ID of the protein to which this signature belongs
        o signature_type : <str>   type of elements in the signature [LOOP, DOMAIN]
        o participantsIDs: <list>  IDs of the participants
        '''

        super(ILXMLProteinSignature, self).__init__(ILXMLParser = ILXMLParser, 
                                                    ID          = ID, 
                                                    Type        = SIGNATURE, 
                                                    parentID    = parentID, 
                                                    parentType  = parentType) # PROTEIN
        self.signature_type = signature_type

        raise NotImplementedError("Not Implemented yet!\n")

    # SETTERS 
    # Not implemented yet. The class is not used yet. 

    # GETTERS
    def get_signatureType(self): return self.signature_type

    # OUTPUT
    def __repr__(self): pass


class ILXMLInteraction(ILXMLObject):
    def __init__(self, ILXMLParser, ID, interactor1_ID=None, interactor2_ID=None, Splus=None, Sminus=None, LpVR=None, LSR=None):
        super(ILXMLInteraction, self).__init__(ILXMLParser, ID, INTERACTION)
        self.interactor1_ID = interactor1_ID
        self.interactor2_ID = interactor2_ID
        self.Splus          = Splus
        self.Sminus         = Sminus
        self.LpVR           = LpVR
        self.LSR            = LSR
        self.reset_RFResults()
        self.reset_signatures()

    # SETTERS (information attributes)
    def set_interactor1_ID(self, i1ID)    : self.interactor1_ID = i1ID
    def set_interactor2_ID(self, i2ID)    : self.interactor2_ID = i2ID
    def set_Splus(self, Splus)            : self.Splus          = Splus
    def set_Sminus(self, Sminus)          : self.Sminus         = Sminus
    def set_LpVR(self, LpVR)              : self.LpVR           = LpVR
    def set_LSR(self, LSR)                : self.LSR            = LSR

    # SETTERS (container attributes)
    def add_RFResult(self, RFResultObject): self.get_RFResults().append(RFResultObject)
    def add_signature(self, IntSignObject): self.get_signatures().append(IntSignObject)

    # SETTERS (container attributes clean up)
    def reset_RFResults(self)             : self.RFResults = []
    def reset_signatures(self)            : self.signatures = []

    # GETTERS (information attributes)
    def get_interactor1_ID(self): return self.interactor1_ID
    def get_interactor2_ID(self): return self.interactor2_ID
    def get_Splus(self)         : return self.Splus
    def get_Sminus(self)        : return self.Sminus
    def get_LpVR(self)          : return self.LpVR
    def get_LSR(self)           : return self.LSR
    def get_RFResults(self)     : return self.RFResults
    def get_signatures(self)    : return self.signatures

    # GETTERS (ID/human readable codes conversion)
    def get_i1name(self): 
        i1_protein = self.get_parser().get_proteins()[self.get_interactor1_ID()]
        if isinstance(i1_protein, ILXMLProtein): return i1_protein.get_name()
        else                                   : return i1_protein

    def get_i2name(self): 
        i2_protein = self.get_parser().get_proteins()[self.get_interactor2_ID()]
        if isinstance(i2_protein, ILXMLProtein): return i2_protein.get_name()
        else                                   : return i2_protein

    def get_name(self)  : return repr((self.get_i1name(), self.get_i2name()))

    # GETTERS (container sub-sets)
    def get_positive_signatures(self): return [ x for x in self.get_signatures() if x.get_refMatrix() == POSITIVE ]
    def get_negative_signatures(self): return [ x for x in self.get_signatures() if x.get_refMatrix() == NEGATIVE ]

    # OUTPUT
    def __repr__(self): 
        return "item = INTERACTION, Interaction = %s, Splus = %s, Sminus = %s, LpVR = %s, LSR = %s\n" %(self.get_name(), self.get_Splus(), 
                                                                                                        self.get_Sminus(), self.get_LpVR(), self.get_LSR())


class ILXMLInteractionRFResult(ILXMLSubObj): 
    def __init__(self, ILXMLParser, parentID, parentType, cost=None, prediction=None, RF_score=None):
        '''
        o cost      : repr(<int>)
        o prediction: <bool>
        o RF_score  : repr(<float>)
        '''
    
        super(ILXMLInteractionRFResult, self).__init__(ILXMLParser = ILXMLParser, 
                                                       ID          = None,
                                                       Type        = RFRESULT, 
                                                       parentID    = parentID, 
                                                       parentType  = parentType) # INTERACTION

        self.cost = cost
        self.prediction = prediction
        self.RF_score = RF_score
        self.precisions = []

    # SETTERS (information attributes)
    def set_cost(self, cost)                : self.cost = cost
    def set_prediction(self, prediction)    : self.prediction = prediction
    def set_RFscore(self, RFscore)          : self.RFscore = RFscore

    # SETTERS (containers)
    def add_precision(self, precisionObject): self.get_precisions().append(precisionObject)

    # GETTERS
    def get_cost(self): return self.cost
    def get_prediction(self): return self.prediction
    def get_RFscore(self): return self.RFscore
    def get_precisions(self): return self.precisions

    # OUTPUT
    def __repr__(self):         
        if   self.get_prediction() is False: prediction = 'NO'
        elif self.get_prediction() is True : prediction = 'YES'
        
        return "\titem = RF RESULT, Prediction = %s, Relative cost = %s, RF score = %s\n" %(prediction, self.get_cost(), self.get_RFscore())


class ILXMLPrecision(object): 
    def __init__(self, UR, precision, error): 
        self.UR        = UR
        self.precision = precision
        self.error     = error

    # SETTERS
    # Not reequired. A "buffer" object is initialized with all its contents. 

    # GETTERS
    def get_UR(self)       : return self.UR
    def get_precision(self): return self.precision
    def get_error(self)    : return self.error

    # OUTPUT
    def __repr__(self): 
        return "\t\titem = PRECISION, UR = %s, precision = %s, error = %s\n" %(self.get_UR(), self.get_precision(), self.get_error()) 


class ILXMLInteractionSignature(ILXMLSubObj): 
    def __init__(self, ILXMLParser, parentID, parentType, refMatrix=None, elementsType=None, P1SignID=None, P2SignID=None, pValue=None):
        super(ILXMLInteractionSignature, self).__init__(ILXMLParser = ILXMLParser, 
                                                        ID          = None,
                                                        Type        = SIGNATURE, 
                                                        parentID    = parentID, 
                                                        parentType  = parentType) # INTERACTION
        self.refMatrix     = refMatrix
        self.elementsType  = elementsType
        self.P1SignID      = P1SignID
        self.P2SignID      = P2SignID
        self.pValue        = pValue

    # SETTERS
    def set_refMatrix(self, refMatrix)      : self.refMatrix    = refMatrix
    def set_elementsType(self, elementsType): self.elementsType = elementsType
    def set_P1SignID(self, P1SignID)        : self.P1SignID     = P1SignID
    def set_P2SignID(self, P2SignID)        : self.P2SignID     = P2SignID
    def set_pValue(self, pValue)            : self.pValue       = pValue

    # GETTERS
    def get_refMatrix(self)   : return self.refMatrix
    def get_elementsType(self): return self.elementsType
    def get_P1SignID(self)    : return self.P1SignID
    def get_P2SignID(self)    : return self.P2SignID
    def get_pValue(self)      : return self.pValue

    # OUTPUT
    def __repr__(self): 
        refMat    = self.get_refMatrix()
        signature = repr((tuple(self.get_parser().signatureID_2_signatureCode(self.get_elementsType(), self.get_P1SignID())), 
                          tuple(self.get_parser().signatureID_2_signatureCode(self.get_elementsType(), self.get_P2SignID()))))
        pValue    = self.get_pValue()

        return "\titem = INTERACTION SIGNATURE, type = %s, signature = %s, pValue = %s\n" %(refMat, signature, pValue)

class IloopsParser(ILXMLParser):

    def custom_interaction_output(self, interaction_object, **kwds): 
        return interaction_object

    def custom_protein_output(self, protein_object, **kwds):
        return protein_object

    def custom_feature_output(self, feature_subObj, feature_type, **kwds):
        return feature_subObj

    def custom_interaction_signature_output(self, int_signature_subObj, signature_type, **kwds):
        return int_signature_subObj

    def set_xml_file(self, xml_file):
        self.xml_file = xml_file

    def get_protein_loops(self, protein_name, warnings=False, **kwds):
        '''
        Get a list of the loops for a certain protein
        '''
        loops = []
        for protein in self.results_parser(xml_file=self.xml_file, report_level=0, **kwds):
            if isinstance(protein, ILXMLProtein) and protein.get_name() == protein_name:
                for feature in protein.get_features():
                    loop_subClass = feature.get_subClass()
                    align_ini = int(feature.get_ali_ini())
                    align_end = int(feature.get_ali_end())
                    loops.append((loop_subClass, align_ini, align_end))
                break
        if len(loops) == 0 and warnings:
            sys.stderr.write('No loops found for protein %s\n' % protein_name)
        return loops
    
    def get_interaction(self, protein1, protein2):
        pass

    def get_positive_signatures(self, protein1, protein2, max_protein_signatures = 3, warnings=False, **kwds):
        if max_protein_signatures > 3:
           max_protein_signatures = 3
        if max_protein_signatures < 1:
           max_protein_signatures = 1
        positive_signatures = []
        for interaction in self.results_parser(xml_file=self.xml_file, report_level=0, **kwds):
            if isinstance(interaction, ILXMLInteraction) and interaction.get_i1name() == protein1 and interaction.get_i2name() == protein2:
                for signature in interaction.get_positive_signatures():
                    P1IS = list(signature.get_parser().signatureID_2_signatureCode(signature.get_elementsType(), signature.get_P1SignID()))
                    P2IS = list(signature.get_parser().signatureID_2_signatureCode(signature.get_elementsType(), signature.get_P2SignID()))
                    pValue = float(signature.get_pValue())
                    if len(P1IS) <= max_protein_signatures and len(P2IS) <= max_protein_signatures:
                        positive_signatures.append((P1IS, P2IS, pValue))
                break
        if len(positive_signatures) == 0 and warnings:
            sys.stderr.write('No positive signatures found for interaction (%s,  %s)\n' % (protein1, protein2))
        return positive_signatures


    def get_negative_signatures(self, protein1, protein2, max_protein_signatures = 3, warnings=False, **kwds):
        if max_protein_signatures > 3:
            max_protein_signatures = 3
        if max_protein_signatures < 1:
           max_protein_signatures = 1
        negative_signatures = []
        for interaction in self.results_parser(xml_file=self.xml_file, report_level=0, **kwds):
            if isinstance(interaction, ILXMLInteraction) and interaction.get_i1name() == protein1 and interaction.get_i2name() == protein2:
                for signature in interaction.get_negative_signatures():
                    P1IS = list(signature.get_parser().signatureID_2_signatureCode(signature.get_elementsType(), signature.get_P1SignID()))
                    P2IS = list(signature.get_parser().signatureID_2_signatureCode(signature.get_elementsType(), signature.get_P2SignID()))
                    pValue = float(signature.get_pValue())
                    if len(P1IS) <= max_protein_signatures and len(P2IS) <= max_protein_signatures:
                        negative_signatures.append((P1IS, P2IS, pValue))
                break
        if len(negative_signatures) == 0 and warnings:
            sys.stderr.write('No negative signatures found for interaction (%s,  %s)\n' % (protein1, protein2))
        return negative_signatures

if __name__ == "__main__":

    import time
    parser = IloopsParser()
    parser.set_xml_file("homodimer.iLoops.59.xml")
    start = time.time()
    loops = parser.get_protein_loops("1tef_A")
    #p_sign = parser.get_positive_signatures("1te0_A", "1te0_B", max_protein_signatures=1)
    #n_sign = parser.get_negative_signatures("NXT1_HUMAN", "Q68CW9_HUMAN", max_protein_signatures=1)
    end=time.time()
    for i in loops:
        print repr(i)
    print end-start



import os, sys, re
import collections
import gzip
import itertools
import json
import numpy
import subprocess
import pwd

# Add "." to sys.path #
sys.path.append(os.path.dirname(__file__))

# Imports jbonet's module #
from SBI.structure       import PDB
from SBI.structure.chain import ChainOfNucleotide
#from SBI.data            import aminoacids3to1, aminoacids_polarity_boolean, aminoacids_surface, nitrogenous_bases, dna_complementary
from SBI.data            import aminoacids3to1, aminoacids_polarity_boolean, aminoacids_surface
from SBI.data            import *

#-------------#
# Classes     #
#-------------#

class DSSP(object):
    '''
    This class defines a DSSP object.
    '''

    def __init__(self, dssp_file):
        self._file = dssp_file
        self._residues = None

    def _check_parsing(self):
        if self._residues == None:
            self._residues = {}
            self._parse_file()

    def _get_file(self):
        return self._file

    def _parse_file(self):
        if os.path.exists(self._get_file()):
            fd = open(self._get_file(), "r")
            for line in fd:
                # Get DSSP info #
                m = re.search("(\d+)\s\S\s[ACDEFGHIKLMNPQRSTVWY]", line)
                if m:
                    chain = line[11:12]
                    residue_num = int(line[5:10])
                    accessible_surface_area = float(line[35:38])
                    secondary_structure = line[16:17]
                    self._residues[(chain, residue_num)] = (accessible_surface_area, secondary_structure)
            fd.close()
        else:
            raise ValueError("Could not open DSSP file %s" % self._get_file())

    def has_residue(self, chain, residue_num):
        self._check_parsing()

        if (chain, residue_num) in self._residues:
            return True

        return False

    def get_accessible_surface_area(self, chain, residue_num):
        self._check_parsing()

        if self.has_residue(chain, residue_num):
            return self._residues[(chain, residue_num)][0]

        return None

    def get_secondary_structure(self, chain, residue_num):
        self._check_parsing()

        secondary_structures = set("EH")

        if self._residues[(chain, residue_num)][1] not in secondary_structures:
            return "C" # residue is in a "coil" region

        return self._residues[(chain, residue_num)][1]

class PirAlignmentMod(object):
    '''
    This class defines a PIR alignment file.
    '''

    def __init__(self, alignment_file):
        self._file = alignment_file
        self._query = None
        self._pdb_name = None
        #self._type = None
        #self._pdb_chains = None
        # self._potential = None
        self._query_alignments = None
        self._hit_alignments = None

        self._parse_file()

    def _get_file(self):
        return self._file

    def _parse_file(self):
        if os.path.exists(self._get_file()):
            self._query_alignments = []
            self._hit_alignments = []
            alignment = None

            fd = open(self._get_file(), "rt")
            for line in fd.readlines():
                # m = re.search("^structure:(.{4}):\d+:.:\d+:.:(dimer|monomer):(\S+):(family|general):", line)
                m = re.search("^structureX:(.{4}):\d+:(\w):.:(\w):", line)
                if m:
                    self._pdb_name = m.group(1)
                    self._pdb_chains = m.group(2), m.group(3)
                    alignment = "structure"
                m = re.search("^sequence:(\S+):", line)
                if m:
                    self._query = m.group(1)
                    alignment = "sequence"
                m = re.search("^(.+)[/*]$", line)
                if m:
                    if alignment == "structure":
                        self._hit_alignments.append(m.group(1))
                    if alignment == "sequence":
                        self._query_alignments.append(m.group(1))
            fd.close()
        else:
            raise ValueError("Could not open statistical potentials file %s" % self._get_file())

    def get_query(self):
        return self._query

    def get_pdb_name(self):
        return self._pdb_name

    def get_pdb_chains(self):
        #return sorted(self._pdb_chains)
        return self._pdb_chains

    def get_type(self):
        return self._type

    # def get_potential(self):
    #     return self._potential

    def get_alignments(self, atype="query"):
        if atype != "query" and atype != "hit":
            atype = "query"

        if atype == "query":
            return self._query_alignments

        return self._hit_alignments

#-------------#
# Accessory   #
#-------------#

def make_subdirs(main, subdirs):
    '''
    This function makes all subdirs listed in "subdirs".
    '''

    for subdir in subdirs:
        if not os.path.exists(os.path.join(main, subdir)):
            os.makedirs(os.path.join(main, subdir))

def remove_files(files):
    '''
    This function removes all files listed in "files".
    '''

    for each_file in files:
        if os.path.exists(each_file):
            os.remove(each_file)

#-------------#
# Parsers     #
#-------------#

def parse_list_file(list_file, gz=False):
    '''
    This function parses any "list" file and returns a generator.
    '''
    #print("Parse list %s\n"%list_file)
    if os.path.exists(list_file):
        fd = None

        if gz: fd = gzip.open(list_file, "rt")
        else:  fd = open(list_file, "rt")
        for line in fd:
            yield line.strip("\n")
        fd.close()
    else:
        raise ValueError("Could not open list file %s" % list_file)

def parse_csv_file(csv_file, gz=False):
    '''
    This function parses any "CSV" file and returns a generator of lists.
    '''

    if os.path.exists(csv_file):
        fd = None

        if gz: fd = gzip.open(csv_file, "rt")
        else:  fd = open(csv_file, "rt")
        for line in fd:
            yield line.strip("\n").split(",")
        fd.close()
    else:
        raise ValueError("Could not open CSV file %s" % csv_file)

def parse_fasta_file(fasta_file, gz=False, clean=True):
    '''
    This function parses any "FASTA" file and iteratively yields sequences as a tuple
    of the form (identifier, sequence).
    '''

    if os.path.exists(fasta_file):
        identifier = ""
        sequence = ""

        for line in parse_list_file(fasta_file, gz):
            line = line.strip()
            if line[0] == ">":
                if sequence != "":
                    yield (identifier, sequence)
                m = re.search("^>(\S+)\|*", line)
                identifier = m.group(1)
                sequence = ""
            else:
                if clean:
                    sequence += re.sub("\W|\d", "X", line.upper())
                else:
                    sequence += line.upper()
        yield (identifier, sequence)
    else:
        raise ValueError("Could not open FASTA file %s" % fasta_file)

#------------#
# SBI module #
#------------#

def get_aminoacid_cb_or_ca(aminoacid):
    """
    Returns the CB {Atom} or the CA {Atom}; otherwise return "None"
    @rtype: {Atom}
    """

    if aminoacid.has_cb:
        return aminoacid.cb
    elif aminoacid.has_ca: # for glycines
        return aminoacid.ca

    return None


#-------------#
# PDB         #
#-------------#

def get_residue_residue_contacts(chain_A, chain_B):
    '''
    This function gets all residue-residue contacts according to the
    definition my Mosca R., Ceol A. & Aloy P., 2013.
    '''
    disulfide_bridge_threshold = 2.56
    hydrogen_bond_threshold = 3.5
    salt_bridge_threshold = 5.5
    van_der_waals_threshold = 5.0
    contacts = {}

    # For each amino acid... #
    for aminoacid_A in chain_A.aminoacids:
        # For each amino acid... #
        for aminoacid_B in chain_B.aminoacids:
            contact = aminoacid_A.distance(aminoacid_B, "min")
            # Skip if amino acids are too apart or distance could not be calcualted #
            if contact[-1] == -1 or contact[-1] > 5.5: continue
            # For each atom... #
            for atom_A in aminoacid_A.atoms:
                atom_type_A = get_atom_type(atom_A.pretty_name)
                # Skip if atom type cannot make contacts #
                if atom_type_A == None: continue
                # For each atom... #
                for atom_B in aminoacid_B.atoms:
                    atom_type_B = get_atom_type(atom_B.pretty_name)
                    # Skip if atom type cannot make contacts #
                    if atom_type_B == None: continue
                    # Get distance #
                    distance = atom_A.distance(atom_B)
                    # Skip if atoms are too apart or distance could not be calcualted #
                    if distance == -1 or distance > 5.5: continue
                    # Get interactions #
                    disulfide_bridge = is_disulfide_bridge(aminoacid_A, atom_type_A, aminoacid_B, atom_type_B, distance)
                    hydrogen_bond = is_hydrogen_bond(aminoacid_A, atom_type_A, aminoacid_B, atom_type_B, distance)
                    salt_bridge = is_salt_bridge(aminoacid_A, atom_type_A, aminoacid_B, atom_type_B, distance)
                    van_der_waals = is_van_der_waals(aminoacid_A, atom_type_A, aminoacid_B, atom_type_B, distance)
                    if disulfide_bridge or hydrogen_bond or salt_bridge or van_der_waals:
                        if atoms_clash(atom_type_A, atom_type_B, distance, disulfide_bridge) == False:
                            contacts.setdefault(((chain_A.chain, aminoacid_A.number), (chain_B.chain, aminoacid_B.number)), [])
                            contacts[(chain_A.chain, aminoacid_A.number), (chain_B.chain, aminoacid_B.number)].append([(aminoacid_A.standard_type, atom_type_A), (aminoacid_B.standard_type, atom_type_B), distance, disulfide_bridge, hydrogen_bond, salt_bridge, van_der_waals])

    return contacts

def get_atom_type(atom_name):
    '''
    This function returns wheter an atom is a "C", "N", "O", "S", or "None" of these.
    '''

    m = re.search("^ C  |\WC\WA|\WC\WB|\WC\WC|\WC\WD|\WC\WE|\WC\WF|\WC\WG|\WC\WH|\WC\WI|\WC\WJ|\WC\WK|\WC\WL|\WC\WM|\WC\WN|\WC\WO|\WC\WP|\WC\WQ|\WC\WR|\WC\WS|\WC\WT|\WC\WU|\WC\WV|\WC\WW|\WC\WX|\WC\WY|\WCA\W|\WCAA|\WCAB|\WCAC|\WCAE|\WCAF|\WCAG|\WCAH|\WCAI|\WCAJ|\WCAK|\WCAL|\WCAM|\WCAN|\WCAO|\WCAP|\WCAQ|\WCAR|\WCAS|\WCAT|\WCAU|\WCAX|\WCAY|\WCAZ|\WCB\W|\WCBA|\WCBB|\WCBC|\WCBD|\WCBE|\WCBF|\WCBG|\WCBH|\WCBI|\WCBX|\WCC\W|\WCCA|\WCCB|\WCCX|\WCD\W|\WCDA|\WCDX|\WCE\W|\WCEA|\WCEB|\WCEC|\WCEX|\WCF\W|\WCFX|\WCG\W|\WCGC|\WCGX|\WCH\W|\WCHA|\WCHB|\WCHC|\WCHD|\WCHX|\WCI\W|\WCIX|\WCJ\W|\WCJX|\WCK\W|\WCKX|\WCL\W|\WCM\W|\WCMA|\WCMB|\WCMC|\WCMD|\WCME|\WCMP|\WCMT|\WCMX|\WCMZ|\WCN\W|\WCNT|\WCNX|\WCO\W|\WCOX|\WCP\W|\WCPX|\WCQ\W|\WCQX|\WCS\W|\WCT\W|\WCW\W|\WCX\W|\WCXD|\WCXE|\WCXF|\WCXG|\WCXN|\WCXO|\WCXP|\WCXQ|\WCXT|\WCXU|\WCXV|\WCYT|\WCZ\W$", atom_name)
    if m:
        return "C"
    m = re.search("^ N  |\WN\WA|\WN\WB|\WN\WC|\WN\WD|\WN\WE|\WN\WF|\WN\WM|\WN\WP|\WN\WS|\WN\WT|\WN\WX|\WN\WY|\WNA\W|\WNAA|\WNAB|\WNAH|\WNAI|\WNAV|\WNAW|\WNAX|\WNB\W|\WNBI|\WNBJ|\WNBK|\WNC\W|\WND\W|\WNE\W|\WNF\W|\WNG\W|\WNH\W|\WNI\W|\WNJ\W|\WNK\W|\WNL\W|\WNLX|\WNM\W|\WNN\W|\WNO\W|\WNP\W|\WNPA|\WNPB|\WNPC|\WNPD|\WNQ\W|\WNR\W|\WNRO|\WNS\W|\WNT\W|\WNW\W|\WNX\W|\WNXT|\WNXU|\WNXV|\WNXW|\WNZ\W$", atom_name)
    if m:
        return "N"
    m = re.search("^ O  |\WO\WA|\WO\WB|\WO\WC|\WO\WD|\WO\WE|\WO\WF|\WO\WG|\WO\WL|\WO\WM|\WO\WP|\WO\WQ|\WO\WR|\WO\WS|\WO\WT|\WO\WV|\WO\WX|\WO\WY|\WOA\W|\WOAB|\WOAD|\WOAP|\WOAX|\WOAY|\WOB\W|\WOBC|\WOC\W|\WOCC|\WOCD|\WOD\W|\WODA|\WODB|\WOE\W|\WOEA|\WOF\W|\WOG\W|\WOGL|\WOH\W|\WOHB|\WOHN|\WOI\W|\WOJ\W|\WOL\W|\WOM\W|\WON\W|\WOO\W|\WOP\W|\WOPP|\WOR\W|\WORA|\WOS\W|\WOT\W|\WOW\W|\WOX\W|\WOXA|\WOXS|\WOXT|\WOXX$", atom_name)
    if m:
        return "O"
    m = re.search("^ S  |\WS\WA|\WS\WG|\WS\WP|\WSBC|\WSC\W|\WSD\W|\WSG\W|\WSP\W$", atom_name)
    if m:
        return "S"

    return None

def is_disulfide_bridge(aminoacid_A, atom_type_A, aminoacid_B, atom_type_B, distance):
    '''
    This function returns whether two atoms form a disulfide bridge or not.
    
    According to definition by Mosca R., Ceol A. & Aloy P., 2013, a disulfide
    bridge is formed by any atom pair S-S from 2 Cys at 2.56A.
    '''

    return aminoacid_A.single_letter == "C" and aminoacid_B.single_letter == "C" and atom_type_A == "S" and atom_type_B == "S" and distance <= 2.56

def is_hydrogen_bond(aminoacid_A, atom_type_A, aminoacid_B, atom_type_B, distance):
    '''
    This function returns whether two atoms form an hydrogen bond or not.
    
    According to definition by Mosca R., Ceol A. & Aloy P., 2013, an hydrogen
    bond is formed by any atom pair N-O and O-N at 3.5A.
    '''


    return ((atom_type_A == "N" and atom_type_B == "O") or (atom_type_A == "O" and atom_type_B == "N")) and distance <= 3.5

def is_salt_bridge(aminoacid_A, atom_type_A, aminoacid_B, atom_type_B, distance):
    '''
    This function returns whether two atoms form a salt bridge or not.
    
    According to definition by Mosca R., Ceol A. & Aloy P., 2013, a salt
    bridge is formed by any atom pair N-O and O-N at 5.5A.
    '''
    negative_residues = ['D', 'E']
    positive_residues = ['R', 'H', 'K', 'S', 'Y']

    if aminoacid_A.single_letter in negative_residues and aminoacid_B.single_letter in positive_residues:
        return atom_type_A == "O" and atom_type_B == "N" and distance <= 5.5

    if aminoacid_A.single_letter in positive_residues and aminoacid_B.single_letter in negative_residues:
        return atom_type_A == "N" and atom_type_B == "O" and distance <= 5.5

    return False

def is_van_der_waals(aminoacid_A, atom_type_A, aminoacid_B, atom_type_B, distance):
    '''
    This function returns whether two atoms form a van der waals interaction
    or not.
    
    According to definition by Mosca R., Ceol A. & Aloy P., 2013, a van der
    waals interaction is formed by any atom pair C-C at 5.0A.
    '''

    return atom_type_A == "C" and atom_type_B == "C" and distance <= 5.0

def atoms_clash(atom_type_A, atom_type_B, distance, disulfide_bridge):
    '''
    This function returns wheter two atoms clash or not.

    According to definition by Mosca R., Ceol A. & Aloy P., 2013, any atom
    pairs at distance less than the sum of the two covalent radii plus 0.5A
    that are not forming a disulfide bridge are considered clashes.
    '''
    covalent_radii = {'C': 0.75, 'N': 0.71, 'O': 0.63}

    if disulfide_bridge == False:
        return distance < (covalent_radii[atom_type_A] + covalent_radii[atom_type_B] + 0.5)

    return False

#-------------#
# Modelling   #
#-------------#

def smith_waterman(A, B, emboss_path):
    '''
    This function aligns a pair of sequences "A" and "B" using water
    from the EMBOSS package.

    @return           = tuple(query alignment, hit alignment, query_start, query_end, hit_start, hit_end)

    type query align. = str
    type hit align.   = str
    type query start  = int
    type query end    = int
    type hit start    = int
    type hit end      = int
    '''
    alignment = []
    positions = []
    

    try:
        process = subprocess.Popen([os.path.join(emboss_path, "water"), "-asequence", A, "-bsequence", B, "-gapopen", "10.0", "-gapextend", "0.5", "-outfile", "stdout"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        for line in process.stdout:
            # Skip commented lines #
            if line.startswith("#"): continue
            # Capture alignment #
            m = re.search("^\S+\s+(\d+)\s+([\w-]+)\s+(\d+)$", line)
            if m:
                alignment.append(m.group(2))
                positions.append([int(m.group(1)), int(m.group(3))])

        query_positions = positions[0::2]
        hit_positions = positions[1::2]

        return("".join(alignment[0::2]), "".join(alignment[1::2]), query_positions[0][0], query_positions[-1][-1], hit_positions[0][0], hit_positions[-1][-1])

    except:
        return(None, None, None, None, None, None)

def get_sequence_to_crystal_correlations(pdb, chain, gapped=True):
    '''
    This function gets the correlation 1 to 1 between the sequence
    and the crystal positions, and returns them.
    
    @return           = correlation, correlation

    type correlation  = dict
    '''
    sequence_to_crystal = {}
    crystal_to_sequence = {}

    # For each protein chain... #
    for protein_chain in pdb.proteins:
        # Skip if not chain of interest #
        if protein_chain.chain == chain:
            # Get protein sequence #
            protein_sequence = protein_chain.gapped_protein_sequence
            if gapped == False:
                protein_sequence = protein_chain.protein_sequence
            # Get protein sequence positions #
            sequence_positions = range(len(protein_sequence))
            # Get crystal positions #
            crystal_positions = [aminoacid.number for aminoacid in protein_chain.aminoacids]
            # For each nucleotide... #
            for aminoacid in protein_sequence:
                if aminoacid == "x":
                    sequence_positions.pop(0)
                    continue
                sequence_to_crystal[sequence_positions[0]] = crystal_positions[0]
                crystal_to_sequence[crystal_positions[0]] = sequence_positions[0]
                # Remove last #
                sequence_positions.pop(0)
                crystal_positions.pop(0)

    return sequence_to_crystal, crystal_to_sequence

def filter_hit(query_alignment, hit_alignment, pdb, contacts, dssp, gaps=True):
    '''
    This function filters a hit if the query does not cover all of the core
    (i.e. region of the hit that contacts the DNA). Additionally, if gaps is
    "False", it filters a hit if either the query or the hit alignments are
    gapped. If gaps is set to "True", it allows insertions/deletions in un-
    structured regions (i.e. "C" secondary structure regions) or up to 1 gap
    in structured regions.

    @return           = boolean
    '''

    # Get sequence/PDB correlation #
    sequence_to_crystal, crystal_to_sequence = get_sequence_to_crystal_correlations(pdb=pdb, chain=pdb.chains[0].chain)
    # Get sequence #
    sequence = pdb.chains[0].gapped_protein_sequence
    # Get core residues #
    core_residues = contacts.get_core_residues()
    # Get secondary structure #
    secondary_structure = ""
    for position in range(len(sequence)):
        if position + 1 in core_residues:
            secondary_structure += "*"
        elif sequence[position] == "x":
            secondary_structure += "C"
        else:
            secondary_structure += dssp.get_secondary_structure(pdb.chains[0].chain, sequence_to_crystal[position])
    # Get start, end hit #
    start = sequence.index(hit_alignment.replace("-", ""))
    end = start + len(hit_alignment.replace("-", "")) - 1
    # Get query secondary structure #
    query_secondary_structure = ""
    positions = range(start, end + 1)
    for i in range(len(hit_alignment)):
        if hit_alignment[i] != "-":
            position = positions.pop(0)
            if query_alignment[i] != "-":
                query_secondary_structure += secondary_structure[position]
            else:
                query_secondary_structure += "-"
        else:
            query_secondary_structure += "-"
    # If contact residues are not conserved, skip #
    if len(re.findall("\*", query_secondary_structure)) != len(re.findall("\*", secondary_structure)):
        return True
    # If insertions/deletions in structured region holding contact residues, skip #
    if re.search("[\*|E]+\-+[\*|E]+", query_secondary_structure) or re.search("[\*|H]+\-+[\*|H]+", query_secondary_structure):
        return True

    return False


def realign(a,b,template):
    """Realign the 'b' gapped sequence which was aligned with 'a' upon a new 'template' of 'a' """
    c=''
    k=0
    j=0
    #print "Original sequence A %s length %d"%(a,len(a))
    #print "Original sequence B %s length %d"%(b,len(b))
    #print "New template for  A %s length %d"%(template,len(template))
    while (k<len(a) or j<len(template)):
        if k<len(a) and j<len(template):
            if (template[j] == a[k] and not template[j] == '-'):
                c+=b[k]
                #print"(%d,%d) %s %s %s"%(j,k,c,a[0:k],template[0:k]) 
                k=k+1
                j=j+1
            elif (template[j] == '-'):
                c+='-'
                #print"(%d,%d) %s %s %s"%(j,k,c,a[0:k],template[0:k]) 
                j=j+1
            elif (a[k] == '-'):
                #print"(%d,%d) %s %s %s"%(j,k,c,a[0:k],template[0:k]) 
                if (k+1<len(a)):
                    print("Skip original sequence known gap %d (.. %s ..)\n"%(k,a[k]))
                else:
                    print("Skip known gap %d (.. %s ..)\n"%(k,a[k]))
                k=k+1
            else:
                #print("Some mistake in (%d,%d) A %s Template %s \n"%(j,k,a[0:k],template[j]))
                c+='-'
                #print"(%d,%d) %s %s %s"%(j,k,c,a[0:k],template[0:k]) 
                k=k+1
                j=j+1
        elif j<len(template):
            c+='-'
            j=j+1 
        elif k<len(a):
            c+=b[k]
            k=k+1
    #print "New alignement of B %s"%c
    return c

def fileExist(file):
    '''
    Check existing files
    '''
    if file is not None:
        return os.path.exists(file) and os.path.isfile(file)
    else:
        return False

def printverbose(f,flag,message):
    '''
    Define verbose to print in file 'f', if 'flag'=True, message given as 'message'
    '''
    if flag: f.write("%s"%(message))

def printfasta (out,name,seq):
    '''
    Print Fasta format
    '''
    out.write(">%s\n"%(name))
    n=0;
    w=""
    for s in seq:
        w += s
        n+=1
        if n == 80:
            out.write("%s\n"%(w))
            n=0
            w=""
    if n>0: out.write("%s\n"%(w))

def PPI_iterator(filename):
    '''
    Iterate on PPIs raw data with alternate names
    '''
    fd=open(filename,"r");
    for line in fd:
        if not line.startswith("#"):
            word=line.strip().split()
            if len(word)>4:
                a=word[0]
                b=word[1]
                alt_a=word[2]
                alt_b=word[3]
                yield (a,b,alt_a,alt_b)
            elif len(word)>1:
                a=word[0]
                b=word[1]
                yield (a,b,"","")
            else:
                continue
    fd.close()

def get_nodes(edges):
    '''
    Gets the list of nodes from a list of edges
    '''
    n=set()
    if len(edges)>0:
        for ppi in edges:
            a,b=ppi
            n.add(a)
            n.add(b)
        if len(n)>0: 
            return n
        else:        
            return None
    else:
        return None



#-------------#
# Cluster     #
#-------------#
def submit_command_to_queue(command, queue=None, max_jobs_in_queue=None, queue_file=None, dummy_dir="/tmp", submit="qsub", qstat="qstat"):
    """
    This function submits any {command} to a cluster {queue}.

    @input:
    command {string}
    queue {string} by default it submits to any queue
    max_jobs_in_queue {int} limits the number of jobs in queue
    queue_file is a file with information specific of the cluster for running a queue

    """
    import hashlib

    if max_jobs_in_queue is not None:
        while number_of_jobs_in_queue(qstat) >= max_jobs_in_queue: time.sleep(5)

    cwd = os.path.join(dummy_dir,"sh")
    if not os.path.exists(cwd): os.makedirs(cwd)
    script= os.path.join(cwd,"submit_"+hashlib.sha224(command).hexdigest()+".sh")
    if queue_file is not None:
      fd=open(script,"w")
      with open(queue_file,"r") as queue_standard:
        data=queue_standard.read()
        fd.write(data)
        fd.write("%s\n\n"%(command))
      fd.close()
      queue_standard.close()
      if queue is not None:
       if  submit=="qsub":
          os.system("%s -q %s %s" % (submit, queue,script))
       elif submit=="sbatch":
          os.system("%s -p %s %s" % (submit, queue,script))
       else:
          os.system("%s %s"% (submit,script))
      else:
        os.system("%s %s"% (submit,script))
    else:
      if queue is not None:
        os.system("echo \"%s\" | %s -q %s" % (command, submit, queue))
      else:
        os.system("echo \"%s\" | %s" % (submit,command))


def number_of_jobs_in_queue(qstat="qstat"):
    """
    This functions returns the number of jobs in queue for a given
    user.

    """

    # Initialize #
    user_name = get_username()

    process = subprocess.check_output([qstat, "-u", user_name])

    return len([line for line in process.split("\n") if user_name in line])


def get_username():
    """
    This functions returns the user name.

    """

    return pwd.getpwuid(os.getuid())[0]


def add_hydrogens(config,path,inp,out,dummy_dir):
    #Initialize
    from SBI.structure import PDB
    import shutil
    src_path = config.get('Paths','modppi_path')
    hbplus = config.get('Paths', 'hbplus_path')
    reduce_exe  =  config.get('Paths', 'reduce_path')
    reduce_db  =  config.get('Paths', 'reduce_db_path')
    relax_exe = config.get('Paths', 'relax_exe')
    hydrogen_type = config.get('Parameters','hydrogens')
    relax = config.get('Parameters','relax')
    cwd = os.getcwd()
    os.chdir(path)
    if fileExist(inp):
     if len(inp.split('.')) > 0: output_hbplus = ".".join(inp.split('.')[:-1]) + ".h"
     else: output_hbplus= inp.strip()+".h"
     if hydrogen_type == "full":
       os.system("%s -Quiet %s -DB %s> %s"%(reduce_exe,inp,reduce_db,output_hbplus))
     else:
       os.system("%s -o %s >& hbplus.log"%(hbplus,inp))
     if relax == "yes":
       sys.stdout.write("\t\t\t-- Relaxing the hydrogen-intermediate model %s (see Rosetta output in relax.log and score.sc)...\n"%output_hbplus)
       os.system("%s -s %s -in:file:fullatom -nstruct 1  -packing:repack_only -relax:jump_move false >& relax.log"%(relax_exe,output_hbplus))
       opt_model=".".join(output_hbplus.split('.')[:-1])+"_0001.pdb"
       old_model=".".join(output_hbplus.split('.')[:-1])+"_non_optimized.pdb"
       shutil.move(output_hbplus,old_model)
       if fileExist(opt_model):
          check_pdb=PDB(opt_model)
          if check_pdb.has_protein:
            check_pdb.clean()
            check_pdb.write(output_hbplus)
            try:
              sys.path.remove(opt_model)
            except:
              sys.stdout.write("\t\t\t-- Keeping old file %s ...\n"%opt_model)
          else:
            shutil.copy(old_model,output_hbplus)
       else:
          shutil.copy(old_model,output_hbplus)
       #clean files
       if fileExist(opt_model): 
          if fileExist(os.path.join(dummy_dir,opt_model)):
             os.remove(opt_model)
          else:
             shutil.move(opt_model,dummy_dir)
       if fileExist(old_model): 
          if fileExist(os.path.join(dummy_dir,old_model)):
             os.remove(old_model)
          else:
             shutil.move(old_model,dummy_dir)
     if not fileExist(output_hbplus):
       raise ValueError("Cannot find file with hydrogen atoms")
     else:
      pdb=PDB(output_hbplus)
      pdb.clean()
      pdb.write(out,force=True)
    os.chdir(cwd)

def renumber_pdb(config,path,pdb_name,sequences,dummy_dir):
    ''' 
     Renumber PDB file located in path folder with the real sequences

     path	Folder where PDB file is located
     pdb 	PDB file
     sequences  dictionary of sequences (of ProteinSequence Class from SeqIO) that define the Aa number
                chain identifier is the key of the dictionary
     dummy_dir  Dummy directory to cerate files

    '''

    #Initialize
    from SBI.structure.chain import Chain
    from SBI.sequence import Sequence
    from SBI.structure import PDB
    from Bio import SeqIO
    from Bio import ExPASy
    from Bio import AlignIO
    from Bio.Align import Applications

    clustal_exe  = os.path.join(config.get('Paths','clustal_path'),'clustalw2')
    name_pdb = ".".join(pdb_name.split('/')[-1].split('.')[:-1])
    new_pdb=PDB()
    pdb_file=os.path.join(path,pdb_name)
    pdb=PDB(pdb_file)
    pdb.clean()
    for chain_id,chain_seq in sequences.iteritems():
       name_chain = name_pdb+"_"+chain_id
       name_seq   = chain_seq.get_identifier()
       pdb_chain  = pdb.get_chain_by_id(chain_id)
       new_chain  = Chain(name_pdb,chain_id)
       #define/create files
       infile =dummy_dir+"/tmp_"+name_chain+"_"+name_seq+".fa"
       outfile=dummy_dir+"/tmp_"+name_chain+"_"+name_seq+".aln"
       dndfile=dummy_dir+"/tmp_"+name_chain+"_"+name_seq+".dnd"
       fd=open(infile,"w")
       fd.write(">{0:s}\n{1:s}\n".format(name_chain,pdb_chain.protein_sequence))
       fd.write(">{0:s}\n{1:s}\n".format(name_seq,chain_seq.get_sequence()))
       fd.close()
       try:
         # run clustalw2
         msa_cline=Applications.ClustalwCommandline(clustal_exe,infile=infile,outfile=outfile)
         child = subprocess.Popen(str(msa_cline),stdout=subprocess.PIPE,stderr=subprocess.PIPE,shell="/bin/bash")
         child.communicate()
         #store alignment in compare
         alignment=AlignIO.read(outfile,'clustal')
         structure=alignment[0].seq
         reference=alignment[1].seq
         try:
           len_3d =len(structure)
           len_ref=len(reference)
         except Exception as e:
           sys.stderr.write("ERROR: %s\n"%e)
           return e
       except Exception as e:
         sys.stderr.write("ERROR: %s\n"%e)
         return e
       #remove temporary fasta and alignment files
       remove_files([infile,outfile,dndfile])
       #mapping of residues to the original sequence
       mapping=create_mapping(pdb_chain.protein_idx.split(";"),structure,reference)
       #fill the new chain with the correct numbering of residues
       for residue in pdb_chain.aminoacids:
          pair = (str(residue.number),residue.version)
          number,version = mapping.get(pair)
          residue.number=number
          residue.version=version
          new_chain.add_residue(residue)
       #fill the new pdb
       new_pdb.add_chain(new_chain)
      
    return new_pdb  

def create_mapping(idx,structure,reference):
    mapped={}
    n=0
    m=0
    jump=0
    abc=" ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
    for i in range(len(reference)):
      idx_reference=m+1
      if structure[i]!="-":
         number =idx[n][0:-1]
         version=idx[n][-1]
         if reference[i] == "-":
           if jump < 53: new_version = abc[jump]
           else: new_version="@"
         else:
           new_version = version
         mapped.setdefault((number,version),(idx_reference,new_version))
         n = n +1 
      if reference[i]!="-":
         jump = 0
         m    = m + 1
      else:
         jump = jump + 1
        
    return mapped 



